(********************************************)
(*                                          *)
(*       Interactive Graphics Package       *)
(*          Copyright D.Bailey 2004         *)
(*                                          *)
(********************************************)





BeginPackage["InteractiveGraphics`",{"JLink`"}] 

Interactive\[Breve]Graphics::usage="Interactive\[Breve]Graphics[g,<options>]
Starts the IGP with a 2D graphical object g. Options include:
MouseClickFunction,MouseMotionFunction,MouseLineSelectionFunction,
TimerFunction,TimerInterval,AnimationFunction,Title,HideNotebook"; 

MouseClickFunction::usage=
    "Option for Interactive\[Breve]Graphics: Specifies a 2-argument function \
called each time the mouse is clicked.";

MouseMotionFunction::usage=
    "Option for Interactive\[Breve]Graphics: Specifies a 2-argument function \
that is called each time the mouse is moved.";

MouseLineSelectionFunction::usage=
    "Option for Interactive\[Breve]Graphics: Specifies a 4-argument function \
that is called after a selection event (shown on the graph as a temporary \
line).";

MouseBoxSelectionFunction::usage=
    "Option for Interactive\[Breve]Graphics: Specifies a 4-argument function \
that is called after a selection event (shown on the graph as a temporary \
rectangle).";

TimerFunction::usage=
    "Option for Interactive\[Breve]Graphics: Specifies a 0-argument function \
that is called at the specified time interval.";

TimerInterval::usage=
    "Option for Interactive\[Breve]Graphics: Specifies the time interval to \
use in seconds (default 10).";

AnimationFunction::usage=
    "Option for Interactive\[Breve]Graphics: Specifies a 0-argument function \
to be called each time a new animation frame is required.";

MenuBar::usage="Option for InteractiveGraphics: Specifies a menu bar for the \
window. A menu is specified as a list {item1,item2,....} where each item can \
be either
a)  {name,function}
b)  {name,sub-menu}
c)  Seperator"

Separator::usage="Used inside menu specifications - see MenuBar"

Title::usage=
    "Option for Interactive\[Breve]Graphics: Specifies a title for the IGP \
window. If anything other than a string is supplied, it will be converted by \
the function ToString[#,StandardForm]&";

HideNotebook::usage="Option for Interactive\[Breve]Graphics: ";

Animation\[Breve]Time::usage="Animation\[Breve]Time[]
Returns the time in seconds since the animation began, or zero.";

Update\[Breve]Graphics::usage="Update\[Breve]Graphics[g,<options>]
Supplies the 2D graphic g to the IGP (which must be displaying a plot). The \
option ReCalibate can be used to specify if re-calibration is required.";

ReCalibate::usage="Option for Update\[Breve]Graphics: Set to True to force a \
ReCalibration (e.g. if the plot axes may have changed)."

Get\[Breve]Mouse\[Breve]Click\[Breve]Identifier::usage="Get\[Breve]Mouse\
\[Breve]Click\[Breve]Identifier[]
Returns an identification code which encodes the type of mouse key pressed \
and the state of the keyboard \[ControlKey], \[AltKey], \[ShiftKey] keys. The \
value is only valid inside a mouse click function."

Set\[Breve]Default\[Breve]Cursor::usage="Set\[Breve]Default\[Breve]Cursor[]
Sets the default (pointer) cursor in the IGP window.";

Set\[Breve]Hand\[Breve]Cursor::usage="Set\[Breve]Hand\[Breve]Cursor[]
Sets the hand cursor (typically used for clickable objects) in the IGP \
window";

Set\[Breve]Wait\[Breve]Cursor::usage="Set\[Breve]Wait\[Breve]Cursor[]
Sets the wait cursor (typically displayed temporarily while a lengthy \
calculation is performed) in the IGP window.";

Close\[Breve]Graphics\[Breve]Window::usage="Closes the IGP window - typically \
called from a menu function"

Begin["`Private`"]

(* Globals: 
      user\[Breve]interacting
    mathCanvas 
      a\[Breve]x b\[Breve]x a\[Breve]y b\[Breve]y 
      start\[Breve]time 
      current\[Breve]plot 
      frame
      pressed\[Breve]x pressed\[Breve]y 
      current\[Breve]x current\[Breve]y 
      selection\[Breve]mode
      press\[Breve]function
      motion\[Breve]function
      selection\[Breve]function
      permitted\[Breve]graphics\[Breve]types
      click\[Breve]mouse\[Breve]id
      gdc
      timer\[Breve]id
      screen\[Breve]height
      screen\[Breve]width
      frame\[Breve]toolkit
  *)
selection\[Breve]mode;
      user\[Breve]interacting=False;
      click\[Breve]mouse\[Breve]id=0;
      permitted\[Breve]graphics\[Breve]types=(_Graphics | _SurfaceGraphics \
|_ContourGraphics | _DensityGraphics);
    IGP::BadGraphics="The IGP can only handle 2D graphics";
    IGP::BadOption="The option `1` is unknown";
    IGP::BadOptionCombination="This combination of options is invalid";
    IGP::InvalidMenu="Invalid menu specified";
    ToExpression[
        "{a\[Breve]x,a\[Breve]y,b\[Breve]x,b\[Breve]y,frame,pressed\[Breve]x,\
pressed\[Breve]y,current\[Breve]x,current\[Breve]y,selection\[Breve]mode,gdc,\
user\[Breve]interacting,Title,title}"];
    
    Interactive\[Breve]Graphics[g_,opts___Rule]:=
      Module[{mml,mm2,mm3,mm4,pts,ww,x1,y1,tmp,f1,f2,hide\[Breve]notebook,
          timer\[Breve]function,timer\[Breve]interval,
          animation\[Breve]function,dt,ti,the\[Breve]menu},
        InstallJava[];
        JavaBlock[
          start\[Breve]time=0;
          selection\[Breve]mode=0;
          
          click\[Breve]function=
            MouseClickFunction/.{opts}/.MouseClickFunction\[Rule]
                Identity\[Breve]Function;
          
          motion\[Breve]function=
            MouseMotionFunction/.{opts}/.MouseMotionFunction\[Rule]
                Identity\[Breve]Function;
          
          f1=MouseLineSelectionFunction/.{opts}/.MouseLineSelectionFunction\
\[Rule]Identity\[Breve]Function;
          
          f2=MouseBoxSelectionFunction/.{opts}/.MouseBoxSelectionFunction\
\[Rule]Identity\[Breve]Function;
          
          timer\[Breve]function=
            TimerFunction/.{opts}/.TimerFunction\[Rule]Identity\[Breve]\
Function;
          
          timer\[Breve]interval=
            TimerInterval/.{opts}/.TimerInterval\[Rule]10;
          
          animation\[Breve]function=
            AnimationFunction/.{opts}/.AnimationFunction\[Rule]Identity\
\[Breve]Function;
          the\[Breve]menu=MenuBar/.{opts}/.MenuBar\[Rule]Null;
          
          If[animation\[Breve]function=!=Identity\[Breve]Function,
            start\[Breve]time=AbsoluteTime[]];
          
          hide\[Breve]notebook=
            HideNotebook/.{opts}/.HideNotebook\[Rule]False;
          If[f1=!=Identity\[Breve]Function,
            selection\[Breve]function=f1;
            selection\[Breve]mode=1;
            
            If[f2=!=Identity\[Breve]Function,
              Message[IGP::BadOptionCombination];Abort[]];
            ];
          If[f2=!=Identity\[Breve]Function,
            selection\[Breve]function=f2;
            selection\[Breve]mode=2;
            ];
          
          tmp=Complement[
              Map[First,{opts}],{MouseClickFunction,MouseMotionFunction,
                MouseLineSelectionFunction,MouseBoxSelectionFunction,Title,
                HideNotebook,TimerFunction,TimerInterval,AnimationFunction,
                MenuBar}];
          If[tmp =!={},Message[IGP::BadOption,First[tmp]];Abort[]]; 
          title=Title/.{opts}/.Title\[Rule]"Interactive Graphics Package";
          If[Head[title]=!=String,title=ToString[title,StandardForm]];
          frame=JavaNew["com.wolfram.jlink.MathFrame"];
          frame@setLayout[JavaNew["java.awt.BorderLayout"]];
          mathCanvas=JavaNew["com.wolfram.jlink.MathCanvas"];
          frame@setTitle[title];
          frame\[Breve]toolkit=frame@getToolkit[];
          screen\[Breve]height=frame\[Breve]toolkit@getScreenSize[]@height;
          screen\[Breve]width=frame\[Breve]toolkit@getScreenSize[]@width;
          LoadJavaClass["java.awt.Cursor"];
          LoadJavaClass["java.awt.Color"];
                        (* mathCanvas@setUsesFE[True]; *)
          mathCanvas@setImageType[MathCanvas`GRAPHICS];
          (*mathCanvas@setUsesFE[True];*)
          If[the\[Breve]menu =!=Null,
            menubar=JavaNew["java.awt.MenuBar"];
            Scan[menubar@add[build\[Breve]menu[#]]&,the\[Breve]menu];
            frame@setMenuBar[menubar];
            ];
          frame@add["Center",mathCanvas];
          
          tmp=ImageSize/.Cases[
                    g,_[ImageSize,_],\[Infinity]]/.ImageSize\[Rule]{400,
                    400}/.Automatic\[Rule]{400,400};
          (* 
            frame@setSize[tmp\[LeftDoubleBracket]1\[RightDoubleBracket],
                tmp\[LeftDoubleBracket]2\[RightDoubleBracket]]; *)
          window\[Breve]width=tmp[[1]];
          window\[Breve]height=tmp[[2]];
          window\[Breve]x=(screen\[Breve]width-window\[Breve]width)/2;
          window\[Breve]y=(screen\[Breve]height-window\[Breve]height)/2;
          
          frame@setBounds[window\[Breve]x,window\[Breve]y,
              window\[Breve]width+50,window\[Breve]height+50];
          
          mathCanvas@
            setSize[tmp\[LeftDoubleBracket]1\[RightDoubleBracket],
              tmp\[LeftDoubleBracket]2\[RightDoubleBracket]];
          frame@validate[];
          JavaShow[frame];
          user\[Breve]interacting=True;
          Calibrate\[Breve]Plot[g];
          
          If[hide\[Breve]notebook,
            SetOptions[EvaluationNotebook[],Visible\[Rule]False]];
          gdc=mathCanvas@getGraphics[];
          Update\[Breve]Graphics[g];
          mml=JavaNew["com.wolfram.jlink.MathMouseListener"];
          
          mml@setHandler["mousePressed",
              "InteractiveGraphics`Private`mousePressedFunc"];
          
          mml@setHandler["mouseReleased",
              "InteractiveGraphics`Private`mouseReleasedFunc"];
          mathCanvas@addMouseListener[mml];
          mm2=JavaNew["com.wolfram.jlink.MathMouseMotionListener"];
          mm2@
            setHandler["mouseMoved",
              "InteractiveGraphics`Private`Mouse\[Breve]Motion\[Breve]\
Handler"]; 
          
          mm2@setHandler["mouseDragged",
              "InteractiveGraphics`Private`Mouse\[Breve]Drag\[Breve]Handler"];\
 
          mathCanvas@addMouseMotionListener[mm2];
          (*mm3=JavaNew["com.wolfram.jlink.MathWindowListener"];
            mm3@setHandler["windowDeactivated","FocusLossHandler"]; 
            frame@addWindowListener[mm3]; *)
          mm4=JavaNew["com.wolfram.jlink.MathComponentListener"];
          
          mm4@setHandler["componentResized",
              "InteractiveGraphics`Private`ResizeHandler"];
          mathCanvas@addComponentListener[mm4];
          (* 
            Either loop to handle animation (We don't use J/Link Periodicals 
                  because the granularity is too coarse) or, 
            simply switch to a modal window for
            non-animated windows *)
          ti=0;
          
          If[timer\[Breve]function =!= Identity\[Breve]Function,
            ti=AbsoluteTime[]+timer\[Breve]interval];
          
          If[animation\[Breve]function =!=Identity\[Breve]Function || 
              timer\[Breve]function =!= Identity\[Breve]Function,
            While[True,
              If[!frame@isVisible[],Break[]];
              
              If[animation\[Breve]function =!=Identity\[Breve]Function,
                animation\[Breve]function[]];
              dt=ti-AbsoluteTime[];
              If[dt\[LessEqual]0 && ti \[NotEqual] 0,
                timer\[Breve]function[];
                ti=AbsoluteTime[]+timer\[Breve]interval;
                dt=ti-AbsoluteTime[];
                 
                If[animation\[Breve]function ===Identity\[Breve]Function,
                  Pause[dt]]; 
                ];
              ],
            frame@setModal[];
             DoModal[]; 
            ];
          ];
        user\[Breve]interacting=False;
        If[hide\[Breve]notebook,
          SetOptions[EvaluationNotebook[],Visible\[Rule]True]];
        ];

Draw\[Breve]XOR\[Breve]Thing[]:=Module[{},
      gdc@setXORMode[Color`WHITE];
      Switch[selection\[Breve]mode,
        (* Line selection *)
        1,
        gdc@
          drawLine[pressed\[Breve]x,pressed\[Breve]y,current\[Breve]x,
            current\[Breve]y],
        (* Box selection *)
        2,
        gdc@
          drawLine[pressed\[Breve]x,pressed\[Breve]y,current\[Breve]x,
            pressed\[Breve]y];
        gdc@
          drawLine[pressed\[Breve]x,current\[Breve]y,current\[Breve]x,
            current\[Breve]y];
        gdc@
          drawLine[pressed\[Breve]x,pressed\[Breve]y,pressed\[Breve]x,
            current\[Breve]y];
        gdc@
          drawLine[current\[Breve]x,pressed\[Breve]y,current\[Breve]x,
            current\[Breve]y]
        ]
      ];

build\[Breve]menu[{name_String,action_}]:=
    Module[{menu\[Breve]item,listener,menu},
      If[Head[action]===List,
        menu=JavaNew["java.awt.Menu",name];
        Scan[
          If[#===Separator,menu@addSeparator[],
              menu@add[build\[Breve]menu[#]]]&,action];
        menu,
        If[Head[action]=!=Symbol,Message[IGP::InvalidMenu];Abort[]];
        menu\[Breve]item=JavaNew["java.awt.MenuItem",name];
        listener=JavaNew["com.wolfram.jlink.MathActionListener"];
        listener@
          setHandler["actionPerformed",
            With[{action=action},ToString[{#};action[]&]]];
        menu\[Breve]item@addActionListener[listener];
        menu\[Breve]item
        ]
      ];
build\[Breve]menu[_]:=(Message[IGP::InvalidMenu];Abort[]);

mousePressedFunc[event_,x_,y_,_]:=Module[{},
      click\[Breve]mouse\[Breve]id=event@getModifiersEx[];
      pressed\[Breve]x=x;
      pressed\[Breve]y=y;
      current\[Breve]x=x;
      current\[Breve]y=y;
      (* click\[Breve]function[a\[Breve]x*x+b\[Breve]x,
            a\[Breve]y*y+b\[Breve]y]; *)
      ];
mouseReleasedFunc[event_,x_,y_,_]:=Module[{},
      Draw\[Breve]XOR\[Breve]Thing[];
      If[selection\[Breve]mode \[NotEqual] 
            0 && (pressed\[Breve]x \[NotEqual] current\[Breve]x || 
              pressed\[Breve]y \[NotEqual] current\[Breve]y),
        selection\[Breve]function[a\[Breve]x*pressed\[Breve]x+b\[Breve]x,
          a\[Breve]y*pressed\[Breve]y+b\[Breve]y,
          a\[Breve]x*current\[Breve]x+b\[Breve]x,
          a\[Breve]y*current\[Breve]y+b\[Breve]y],
        click\[Breve]function[a\[Breve]x*x+b\[Breve]x,
          a\[Breve]y*y+b\[Breve]y]
        ]
      ];

Mouse\[Breve]Motion\[Breve]Handler[event_,x_,y_,_]:=Module[{},
      motion\[Breve]function[a\[Breve]x*x+b\[Breve]x,
          a\[Breve]y*y+b\[Breve]y];
      ];

ResizeHandler[_]:=
    Update\[Breve]Graphics[current\[Breve]graphics,ReCalibrate\[Rule]True];

Mouse\[Breve]Drag\[Breve]Handler[_,x_,y_,_]:=Module[{g},
      If[pressed\[Breve]x \[NotEqual] current\[Breve]x || 
          pressed\[Breve]y \[NotEqual] current\[Breve]y,
        Draw\[Breve]XOR\[Breve]Thing[]];
      current\[Breve]x=x;
      current\[Breve]y=y;
      Draw\[Breve]XOR\[Breve]Thing[]
      ];

Mouse\[Breve]Motion\[Breve]Handler[_,x_,y_,_]:=Module[{},
      motion\[Breve]function[a\[Breve]x*x+b\[Breve]x,
          a\[Breve]y*y+b\[Breve]y]; 
      ];

Calibrate\[Breve]Plot[plot_]:=Module[{ww,pt1,pt2,pp,nx,ny,pix,c1,c2,epi},
      JavaBlock[
          If[!MatchQ[plot,permitted\[Breve]graphics\[Breve]types],
            Message[IGP::BadGraphics];Abort[]];
          
          ww=AbsoluteOptions[plot,PlotRange]\[LeftDoubleBracket]1,
              2\[RightDoubleBracket];
          {pt1,
              pt2}={{ww\[LeftDoubleBracket]1,1\[RightDoubleBracket],
                ww\[LeftDoubleBracket]2,
                  1\[RightDoubleBracket]},{ww\[LeftDoubleBracket]1,
                  2\[RightDoubleBracket],
                ww\[LeftDoubleBracket]2,2\[RightDoubleBracket]}};
          {pt1,pt2}={0.8*pt1+0.2*pt2,0.2*pt1+0.8*pt2};
          (* 
            Change any bright green or cyan a bit so it does not prevent the \
calibration, 
            and remove any ImageSize information as it can confuse the \
mathCanvas *)
          
          pp=plot/.{RGBColor[0,x_/;x>0.70,0]\[Rule]RGBColor[0,0.70,0],
                RGBColor[0,x_/;x>0.70,y_/;y>0.70]\[Rule]
                  RGBColor[0,0.70,0.70],
                Hue[h_]\[Rule]Hue[h,1,0.70],
                Hue[h_,a_,b_]\[Rule]Hue[h,a,Min[b,0.70]],
                Rule[ImageSize,_]\[RuleDelayed]Sequence[]};
          (* Add a green point and a cyan point *)
          epi={};
          
          pp=pp/.(Epilog\[Rule]x_)\[RuleDelayed](epi=x;kludge)/.kludge\[Rule]
                Sequence[];
          
          pp=Append[pp,
              Epilog\[Rule]{epi,RGBColor[0,1,0],Point[pt1],RGBColor[0,1,1],
                  Point[pt2]}];
          (* 
            pp\[LeftDoubleBracket]1\[RightDoubleBracket]={pp\
\[LeftDoubleBracket]1\[RightDoubleBracket],RGBColor[0,1,0],Point[pt1],
                RGBColor[0,1,1],Point[pt2]}; *)
          mathCanvas@setMathCommand[ToString[pp,InputForm]];
          nx=mathCanvas@getImage[]@getWidth[];
          ny=mathCanvas@getImage[]@getHeight[];
          
          pix=mathCanvas@
              getImage[]@getBufferedImage[]@getRGB[0,0,nx,ny,Null,0,nx];
          pix=Partition[pix,nx];
          (* Green point *)
          c1=Position[pix,-16711936];
          c1=Plus@@c1/Length[c1];
          c1=Reverse[c1];
          (* Cyan point *)
          c2=Position[pix,-16711681];
          c2=Plus@@c2/Length[c2];
          c2=Reverse[c2];
          
          a\[Breve]x=(pt2\[LeftDoubleBracket]1\[RightDoubleBracket]-
                  pt1\[LeftDoubleBracket]1\[RightDoubleBracket])/(c2\
\[LeftDoubleBracket]1\[RightDoubleBracket]-
                  c1\[LeftDoubleBracket]1\[RightDoubleBracket]);
          
          b\[Breve]x=
            pt2\[LeftDoubleBracket]1\[RightDoubleBracket]-
              c2\[LeftDoubleBracket]1\[RightDoubleBracket]*a\[Breve]x;
          
          a\[Breve]y=(pt2\[LeftDoubleBracket]2\[RightDoubleBracket]-
                  pt1\[LeftDoubleBracket]2\[RightDoubleBracket])/(c2\
\[LeftDoubleBracket]2\[RightDoubleBracket]-
                  c1\[LeftDoubleBracket]2\[RightDoubleBracket]);
          
          b\[Breve]y=
            pt2\[LeftDoubleBracket]2\[RightDoubleBracket]-
              c2\[LeftDoubleBracket]2\[RightDoubleBracket]*a\[Breve]y; 
          ];
      ];


Update\[Breve]Graphics[g_,opts___Rule]:=Module[{},
      JavaBlock[
        If[!MatchQ[g,permitted\[Breve]graphics\[Breve]types],
          Message[IGP::BadGraphics];Abort[]];
        current\[Breve]graphics=g;
        If[ReCalibrate/.{opts},Calibrate\[Breve]Plot[g]];
        mathCanvas@
          setMathCommand[
            ToString[g/.{Rule[ImageSize,_]\[RuleDelayed]Sequence[]},
              InputForm]];
        ];
      ];

Close\[Breve]Graphics\[Breve]Window[]:=(frame@dispose[];EndModal[]);

Animation\[Breve]Time[]:=
    If[start\[Breve]time \[NotEqual] 0,AbsoluteTime[]-start\[Breve]time,0];

Get\[Breve]Mouse\[Breve]Click\[Breve]Identifier[]:=
    click\[Breve]mouse\[Breve]id;
Set\[Breve]Default\[Breve]Cursor[]:=
    If[user\[Breve]interacting,
      mathCanvas@
        setCursor[JavaNew["java.awt.Cursor",Cursor`DEFAULTUCURSOR]]];
Set\[Breve]Hand\[Breve]Cursor[]:=
    If[user\[Breve]interacting,
      mathCanvas@setCursor[JavaNew["java.awt.Cursor",Cursor`HANDUCURSOR]]];
Set\[Breve]Wait\[Breve]Cursor[]:=
    If[user\[Breve]interacting,
      mathCanvas@setCursor[JavaNew["java.awt.Cursor",Cursor`WAITUCURSOR]]];

End[]

EndPackage[]