(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.0'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     76766,       1965]*)
(*NotebookOutlinePosition[     84286,       2200]*)
(*  CellTagsIndexPosition[     83788,       2178]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell[TextData[StyleBox["The Interactive Graphics Package",
  FontColor->RGBColor[0, 0, 1]]], "Chapter"],

Cell[CellGroupData[{

Cell["Background information", "Section",
  PageWidth->WindowWidth,
  CellTags->"IGP_1"],

Cell[CellGroupData[{

Cell["Introduction to the package", "Subsection",
  CellTags->"IGP_1"],

Cell[TextData[{
  "This package is about genuine interactive ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " graphics! ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " lets you create superb graphics, but in essence, the graphics system is \
built around the concept that you create one static plot at a time and then \
enjoy it. This is not surprising \[Dash] when the definitions of the relevant \
functions were laid down, there simply was not enough computer resource on \
most machines to consider animated graphics, or graphics that respond to \
mouse clicks. Obviously, it is possible to create an animated plot by \
creating a list of graphics objects (say a list of plots), but this does not \
begin to tap the possibilities that exist for true interactive graphics.\nIn \
more recent times, we have seen the introduction of J/Link, and, even more \
recently, the GUIKit. J/Link lets you create, and control Java applications \
from within ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ". Using J/Link you can create and run interactive graphics applications, \
but you have to create your image using Java primitives \[Dash] leveraging \
none of the power of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " graphics! Alternatively, you can put ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " graphics objects into a Java window, but then you can't interact with \
them! It is possible to attach a \"mouse listener\" to your Java window, and \
actually write ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " code to respond to mouse clicks \[Dash] yet there is no easy way to \
relate the mouse positions to the original graphics coordinates! Rather \
surprisingly, the GUIKit design does not seem to have addressed this problem, \
at least so far. There is also an excellent package, ",
  ButtonBox["MathGL3D",
    ButtonData:>{
      URL[ "http://library.wolfram.com/infocenter/MathSource/2986/"], None},
    ButtonStyle->"Hyperlink"],
  " which will animate 3D objects \[Dash] allowing you to rotate a \
3-dimentional object dynamically using the mouse. However, once again, there \
is no way of specifying an arbitrary interaction with the image \[Dash] say \
to use the mouse to select data points on a plot and cause those points to be \
included/excluded from a curve fit.\nThe Interactive Graphics Package (IGP) \
is an attempt to bring really easy-to-use interactive graphics to ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ". Although the IGP is built upon J/Link, you can use it without any \
knowledge of Java \[Dash] indeed, this is the last you will read of the \
subject in this manual! Currently, the IGP only works in 2 dimensions \[Dash] \
since there is no way to select three coordinates with a mouse!\nThis first \
version of the IGP is extremely simple \[Dash] yet it is already possible to \
construct rather sophisticated programs with it, as I will show. Additional \
features will be added as demand dictates \[Dash] please contact ",
  ButtonBox["the author",
    ButtonData:>{
      URL[ "mailto:dave@dbailey.co.uk"], None},
    ButtonStyle->"Hyperlink"],
  "."
}], "Text",
  CellTags->"IGP_1"],

Cell[TextData[StyleBox["",
  FontSlant->"Italic"]], "Text",
  PageWidth->WindowWidth]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Getting started with the IGP", "Section",
  CellTags->"IGP_3"],

Cell[CellGroupData[{

Cell["Installing the package", "Subsubsection",
  CellTags->"IGP_4"],

Cell[TextData[{
  "The IGP is supplied as a ZIP file containing all the files required in \
their appropriate directories.. It is vital that this directory structure is \
preserved. Copy the file to the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " directory, e.g. C:\\program files\\wolfram research\\mathematica\\5.0 \
(the exact directory depends on the version of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " installed). Unzip the file InteractiveGraphics.ZIP using a tool that \
preserves the directory structure and handles long names correctly, e.g. \
PKZIP(R)  Version 2.50, or WinZip(R).\nFinally, start ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ", click on the 'Help' menu, and select 'Rebuild Help index'. This will \
integrate the IGP documentation with the rest of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ". You will find the IGP help in the 'Addons' section of the help browser."
}], "Text",
  CellTags->"IGP_4"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Loading the package", "Subsubsection",
  CellTags->"IGP_5"],

Cell["The interactive graphics package may be loaded thus:", "Text",
  CellTags->"IGP_5"],

Cell[BoxData[
    RowBox[{"<<", "\"\<InteractiveGraphics`\>\""}]], "Input",
  CellLabel->"In[1]:=",
  CellTags->"IGP_5"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Simple use", "Subsubsection",
  CellTags->"IGP_6"],

Cell["First we load the IGP if necessary:", "Text",
  CellTags->"IGP_6"],

Cell[BoxData[
    RowBox[{"Needs", "[", "\"\<InteractiveGraphics`\>\"", "]"}]], "Input",
  CellLabel->"In[1]:=",
  CellTags->"IGP_6"],

Cell[TextData[{
  "Here is a very simple program which plots a number of points and then lets \
you remove any data points that you wish. Note the use of the ",
  StyleBox["PlotRange",
    FontWeight->"Bold"],
  " option to ensure that the graph always uses the same scale. We use \
DisplayFunction->Identity to stop ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " drawing in the notebook. Remember that this plotting function will be \
called repeatedly as you update the plot interactively \[Dash] so you \
certainly don't want it to produce notebook graphical output."
}], "Text",
  CellTags->"IGP_6"],

Cell["\<\
Define a list of {x,y} points that correspond to a noisy straight line:\
\>", "Text",
  CellTags->"IGP_6"],

Cell[BoxData[
    RowBox[{
      RowBox[{"myData", "=", 
        RowBox[{"Table", "[", 
          RowBox[{
            RowBox[{"{", 
              RowBox[{"x", ",", 
                RowBox[{"x", "+", 
                  RowBox[{"2", "*", 
                    RowBox[{"Random", "[", "]"}]}]}]}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"x", ",", "1", ",", "10"}], "}"}]}], "]"}]}], 
      ";"}]], "Input",
  CellLabel->"In[16]:=",
  CellTags->"IGP_6"],

Cell["Define a function that will plot the data", "Text",
  CellTags->"IGP_6"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"myPlot", "[", "]"}], ":=", 
        RowBox[{"ListPlot", "[", 
          RowBox[{"myData", ",", 
            RowBox[{"PlotRange", "\[Rule]", 
              RowBox[{"{", 
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"0", ",", "10"}], "}"}], ",", 
                  RowBox[{"{", 
                    RowBox[{
                      RowBox[{"-", "2"}], ",", "15"}], "}"}]}], "}"}]}], ",", 
            
            RowBox[{"DisplayFunction", "\[Rule]", "Identity"}], ",", 
            RowBox[{"ImageSize", "\[Rule]", 
              RowBox[{"{", 
                RowBox[{"600", ",", "600"}], "}"}]}]}], "]"}]}], 
      ";"}]], "Input",
  CellLabel->"In[25]:=",
  CellTags->"IGP_6"],

Cell["\<\
This function determines if any particular {x,y}  point is close enough to a \
given data point \[Dash] remember, mouse operations are hardly ever exact.\
\>", "Text",
  CellTags->"IGP_6"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"myClickCriterion", "[", 
          RowBox[{
            RowBox[{"{", 
              RowBox[{"a_", ",", "b_"}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"c_", ",", "d_"}], "}"}]}], "]"}], ":=", 
        RowBox[{"(", 
          RowBox[{
            RowBox[{
              RowBox[{
                RowBox[{"(", 
                  RowBox[{"a", "-", "c"}], ")"}], "^", "2"}], "+", 
              RowBox[{
                RowBox[{"(", 
                  RowBox[{"b", "-", "d"}], ")"}], "^", "2"}]}], " ", "<", " ",
             "0.1"}], ")"}]}], ";"}]], "Input",
  CellLabel->"In[18]:=",
  CellTags->"IGP_6"],

Cell["\<\
This function will be called each time the mouse is pressed (the left button, \
if the mouse has several). It determines if the click is located close enough \
to a data point and if it is, it removes the point and re-draws the graphic. \
Note that we do not generate a diagnostic if the user clicks in the wrong \
place \[Dash] most users are very careless with the mouse!\
\>", "Text",
  CellTags->"IGP_6"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"mouseUpdate", "[", 
          RowBox[{"x_", ",", "y_"}], "]"}], ":=", 
        RowBox[{"Module", "[", 
          RowBox[{
            RowBox[{"{", "pt", "}"}], ",", "\[IndentingNewLine]", 
            RowBox[{
              RowBox[{"pt", "=", 
                RowBox[{"Select", "[", 
                  RowBox[{"myData", ",", 
                    RowBox[{
                      RowBox[{"myClickCriterion", "[", 
                        RowBox[{
                          RowBox[{"{", 
                            RowBox[{"x", ",", "y"}], "}"}], ",", "#"}], "]"}],
                       "&"}]}], "]"}]}], ";", "\[IndentingNewLine]", 
              RowBox[{"If", "[", 
                RowBox[{
                  RowBox[{"pt", " ", "\[NotEqual]", " ", 
                    RowBox[{"{", "}"}]}], ",", "\[IndentingNewLine]", 
                  RowBox[{
                    RowBox[{"pt", "=", 
                      RowBox[{"pt", "[", 
                        RowBox[{"[", "1", "]"}], "]"}]}], ";", 
                    "\[IndentingNewLine]", 
                    RowBox[{"Print", "[", 
                      RowBox[{"\"\<Deleting data point: \>\"", ",", "pt"}], 
                      "]"}], ";", "\[IndentingNewLine]", 
                    RowBox[{"myData", "=", 
                      RowBox[{"DeleteCases", "[", 
                        RowBox[{"myData", ",", "pt"}], "]"}]}], ";", 
                    "\[IndentingNewLine]", 
                    RowBox[{"Update\[Breve]Graphics", "[", 
                      RowBox[{"myPlot", "[", "]"}], "]"}], ";"}]}], 
                "\[IndentingNewLine]", "]"}]}]}], "\[IndentingNewLine]", 
          "]"}]}], ";"}]], "Input",
  CellLabel->"In[19]:=",
  CellTags->"IGP_6"],

Cell[TextData[{
  "Now we invoke the IGP package and print out the resultant data vector \
after the user has closed the window. When you execute this step you should \
see the ",
  StyleBox["ListPlot",
    FontWeight->"Bold"],
  " in a window and you can remove data points simply by clicking on them \
with the mouse."
}], "Text",
  CellTags->"IGP_6"],

Cell[CellGroupData[{

Cell[BoxData[{
    RowBox[{
      RowBox[{"Interactive\[Breve]Graphics", "[", 
        RowBox[{
          RowBox[{"myPlot", "[", "]"}], ",", 
          RowBox[{"MouseClickFunction", "->", "mouseUpdate"}], ",", 
          RowBox[{"Title", "->", "\"\<Data selection window\>\""}]}], "]"}], 
      ";"}], "\[IndentingNewLine]", 
    RowBox[{
      RowBox[{"Print", "[", 
        RowBox[{"\"\<Remaining data=\>\"", ",", "myData"}], "]"}], 
      ";"}]}], "Input",
  CellLabel->"In[26]:=",
  CellTags->"IGP_6"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Deleting data point: \"\>", "\[InvisibleSpace]", 
        RowBox[{"{", 
          RowBox[{"9", ",", "10.143647957517214`"}], "}"}]}],
      SequenceForm[ "Deleting data point: ", {9, 10.143647957517214}],
      Editable->False]], "Print",
  CellLabel->"From In[26]:=",
  CellTags->"IGP_6"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Remaining data=\"\>", "\[InvisibleSpace]", 
        RowBox[{"{", 
          RowBox[{
            RowBox[{"{", 
              RowBox[{"1", ",", "1.646974726456499`"}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"2", ",", "2.363842220408108`"}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"3", ",", "3.13936763654869`"}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"4", ",", "4.423539780042864`"}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"5", ",", "5.2578728210717145`"}], "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"10", ",", "11.647520066452149`"}], "}"}]}], "}"}]}],
      SequenceForm[ 
      "Remaining data=", {{1, 1.6469747264564989}, {2, 2.3638422204081082}, {
        3, 3.1393676365486902}, {4, 4.423539780042864}, {5, 
        5.2578728210717145}, {10, 11.647520066452149}}],
      Editable->False]], "Print",
  CellLabel->"From In[26]:=",
  CellTags->"IGP_6"]
}, Open  ]],

Cell["\<\
It is also possible to respond to the mouse position as the user moves the \
muse across the window. Typically, this is used to adjust the shape of the \
mouse cursor for certain parts of a window.This can create the effect of \
simple buttons, as the next example illustrates. Note that it is extremely \
important not to perform too much processing in response to a mere mouse \
movement, otherwise you may make the mouse move jerkily. Also, it is \
important to realise that your program should not do anything irreversible in \
response to a mouse movement \[Dash] your users will definitely not expect \
that. However, the IGP does not itself impose any constraint on the use of \
this feature.\
\>", "Text",
  CellTags->"IGP_6"],

Cell[TextData[{
  "Here is a summary of the options available to the ",
  StyleBox["Interactive\[Breve]Graphics",
    FontWeight->"Bold"],
  " function. Where an option is used to supply a function, the third column \
indicated the arguments (of real type) that the function should accept."
}], "Text",
  CellTags->"IGP_6"],

Cell[BoxData[GridBox[{
        {
          StyleBox[
            RowBox[{"Option", " ", "to", " ", "Interactive\[Breve]Graphics"}],
            
            "SBO"], 
          StyleBox["Explanation",
            "SBO"], 
          StyleBox[
            RowBox[{"Arguments", " ", "to", " ", "user", " ", "function"}],
            "SBO"]},
        {
          StyleBox[
            RowBox[{"Animation", 
              StyleBox["Function",
                "MR"]}]], 
          RowBox[{
            RowBox[{
            "Function", " ", "called", " ", "when", " ", "a", " ", "new", " ",
               "frame", " ", "of", " ", "the"}], " ", "\[IndentingNewLine]", 
            RowBox[{"animation", " ", "is", " ", 
              RowBox[{"required", ".", " ", 
                StyleBox[Cell[""],
                  "MR"]}]}]}], 
          RowBox[{"[", "]"}]},
        {"MenuBar", 
          RowBox[{
            RowBox[{
            "Supplies", " ", "a", " ", "list", " ", "of", " ", "menu", " ", 
              "items", " ", "which", " ", "form"}], "\[IndentingNewLine]", 
            RowBox[{
            "a", " ", "menu", " ", "that", " ", "is", " ", "attached", " ", 
              "to", " ", "the", " ", 
              RowBox[{"window", "."}]}]}], "\[Dash]"},
        {
          StyleBox["MouseClickFunction",
            "MR"], 
          RowBox[{
            RowBox[{
            "Function", " ", "called", " ", "when", " ", "the", " ", "user", 
              " ", "has", " ", "pressed"}], " ", "\[IndentingNewLine]", 
            RowBox[{"the", " ", "mouse", " ", "and", " ", "released", " ", 
              RowBox[{"(", "clicked", ")"}], " ", 
              RowBox[{"it", ".", " ", 
                StyleBox[Cell[""],
                  "MR"]}]}]}], 
          RowBox[{"[", 
            RowBox[{"x1", ",", "x2"}], "]"}]},
        {
          StyleBox["MouseMotionFunction",
            "MR"], 
          RowBox[{
            RowBox[{
            "Function", " ", "called", " ", "when", " ", "the", " ", "user", 
              " ", "has", " ", "moved"}], " ", "\[IndentingNewLine]", 
            RowBox[{"the", " ", "mouse"}]}], 
          RowBox[{"[", 
            RowBox[{"x1", ",", "x2"}], "]"}]},
        {
          StyleBox["MouseLineSelectionFunction",
            "MR"], 
          RowBox[{
            RowBox[{
            "Function", " ", "called", " ", "when", " ", "the", " ", "user", 
              " ", "has"}], " ", "\[IndentingNewLine]", 
            RowBox[{
              RowBox[{"depressed", " ", "the", " ", "mouse"}], ",", " ", 
              RowBox[{"dragged", " ", "it"}], ",", " ", "and"}], " ", 
            "\[IndentingNewLine]", 
            RowBox[{"released", " ", 
              RowBox[{"it", ".", " ", "A"}], " ", "temporary", " ", "line", 
              " ", 
              StyleBox["is",
                "MR"]}], 
            StyleBox[" ",
              "MR"], 
            StyleBox["\[IndentingNewLine]",
              "MR"], 
            StyleBox[
              RowBox[{
              "displayed", " ", "during", " ", "this", " ", "process"}],
              "MR"], 
            StyleBox["\[IndentingNewLine]",
              "MR"], 
            RowBox[{"as", " ", "a", " ", 
              RowBox[{"guide", ".", 
                StyleBox["                         ",
                  "MR"]}]}]}], 
          RowBox[{"[", 
            RowBox[{"x1", ",", "y1", ",", "x2", ",", "y2"}], "]"}]},
        {
          StyleBox["MouseBoxSelectionFunction",
            "MR"], 
          RowBox[{
            RowBox[{
            "Function", " ", "called", " ", "when", " ", "the", " ", "user", 
              " ", "has"}], " ", "\[IndentingNewLine]", 
            RowBox[{
              RowBox[{"depressed", " ", "the", " ", "mouse"}], ",", " ", 
              RowBox[{"dragged", " ", "it"}], ",", " ", "and"}], " ", 
            "\[IndentingNewLine]", 
            RowBox[{"released", " ", 
              RowBox[{"it", ".", " ", "A"}], " ", "temporary", " ", 
              "rectangular"}], 
            StyleBox["\n",
              "MR"], 
            StyleBox[
              RowBox[{
              "box", " ", "is", " ", "displayed", " ", "during", " ", "this", 
                " ", "process"}],
              "MR"], 
            StyleBox["\[IndentingNewLine]",
              "MR"], 
            RowBox[{"as", " ", "a", " ", 
              RowBox[{"guide", ".", 
                StyleBox["                         ",
                  "MR"]}]}]}], 
          RowBox[{"[", 
            RowBox[{"x1", ",", "y1", ",", "x2", ",", "y2"}], "]"}]},
        {
          StyleBox["HideNotebook",
            "MR"], 
          RowBox[{
            RowBox[{
            "Set", " ", "true", " ", "to", " ", "hide", " ", "the", " ", 
              "notebook", " ", "while", " ", "the"}], " ", 
            "\[IndentingNewLine]", 
            RowBox[{"graphics", " ", "window", " ", "is", " ", 
              RowBox[{"displayed", ".", 
                StyleBox["                         ",
                  "MR"]}]}]}], "\[Dash]"},
        {
          StyleBox["TimerFunction",
            "MR"], 
          RowBox[{
          "Function", " ", "called", " ", "at", " ", "required", " ", "time", 
            " ", 
            RowBox[{"interval", "."}]}], 
          RowBox[{"[", "]"}]},
        {
          StyleBox["TimerInterval",
            "MR"], 
          RowBox[{
          "Interval", " ", "in", " ", "seconds", " ", "between", " ", "timer",
             " ", 
            RowBox[{"calls", "."}]}], "\[Dash]"},
        {
          StyleBox["Title",
            "MR"], 
          RowBox[{
            RowBox[{
            "Supply", " ", "a", " ", "string", " ", "to", " ", "act", " ", 
              "as", " ", "a", " ", "window"}], "\[IndentingNewLine]", 
            RowBox[{"title", ".", 
              StyleBox["                         ",
                "MR"]}]}], "\[Dash]"}
        },
      ColumnAlignments->{Left}]], "Text",
  Background->RGBColor[1, 0.8, 0.8],
  CellTags->"IGP_6"],

Cell[TextData[{
  "You can use Graphics2D,  ContourGraphics, or Density Graphics with the ",
  StyleBox["Interactive\[Breve]Graphics",
    FontWeight->"Bold"],
  " function. Other forms of graphics will give a diagnostic."
}], "Text"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Animated plots", "Section",
  CellTags->"IGP_7"],

Cell[CellGroupData[{

Cell["Introduction", "Subsubsection",
  CellTags->"IGP_7a"],

Cell[TextData[{
  StyleBox["Important note: Prior to version 5.1, animated graphics \
occasionally hang ",
    FontColor->RGBColor[1, 0, 0]],
  StyleBox["Mathematica",
    FontSlant->"Italic",
    FontColor->RGBColor[1, 0, 0]],
  StyleBox[". This seems to be a bug in ",
    FontColor->RGBColor[1, 0, 0]],
  StyleBox["Mathematica",
    FontSlant->"Italic",
    FontColor->RGBColor[1, 0, 0]],
  StyleBox[", and is fixed in version 5.1 \[Dash] if you want to use \
animation it is recommended that you upgrade to that version. ",
    FontColor->RGBColor[1, 0, 0]],
  "\nIf the option AnimationFunction is specified, as soon as a frame has \
been rendered, this function is called (over and over). Typically this \
function calls Animation\[Breve]Time[] to obtain the time in seconds since \
the start of the animation, and uses this to create another frame of the \
animation which it passes to ",
  StyleBox["Update\[Breve]Graphics",
    FontWeight->"Bold"],
  ". The frequency with which this function is called will depend on the \
size/complexity of the image, the speed of the processor, and what other \
tasks are running on the machine. The idea is that because each frame is \
based on the current time into the animation, animations will run at the same \
speed regardless of processor, etc. Faster processors will produce a smoother \
animation. For convenience, the function ",
  StyleBox["Animation\[Breve]Time",
    FontWeight->"Bold"],
  " will return zero if called before an animation has begun. Although this \
can be combined with interaction, it should not be used with the ",
  StyleBox["MouseLineSelectionFunction",
    FontWeight->"Bold"],
  ", or ",
  StyleBox["MouseBoxSelectionFunction",
    FontWeight->"Bold"],
  " options in the current version. Here is a trivial example:"
}], "Text",
  CellTags->"IGP_7a"],

Cell[BoxData[
    RowBox[{"Needs", "[", "\"\<InteractiveGraphics`\>\"", "]"}]], "Input",
  CellLabel->"In[5]:=",
  CellTags->"IGP_6"],

Cell[BoxData[{
    RowBox[{
      RowBox[{
        RowBox[{"myPlot", "[", "]"}], ":=", 
        RowBox[{"Module", "[", 
          RowBox[{
            RowBox[{"{", 
              RowBox[{"tt", "=", 
                RowBox[{"Animation\[Breve]Time", "[", "]"}]}], "}"}], ",", 
            "\[IndentingNewLine]", 
            RowBox[{"Graphics", "[", 
              RowBox[{
                RowBox[{"{", 
                  RowBox[{
                    RowBox[{"Thickness", "[", "0.01", "]"}], ",", 
                    "\[IndentingNewLine]", 
                    RowBox[{"RGBColor", "[", 
                      RowBox[{"1", ",", "0", ",", "0"}], "]"}], ",", 
                    RowBox[{"Circle", "[", 
                      RowBox[{
                        RowBox[{"{", 
                          RowBox[{"0", ",", "0"}], "}"}], ",", "5"}], "]"}], 
                    ",", "\[IndentingNewLine]", 
                    RowBox[{"RGBColor", "[", 
                      RowBox[{"0", ",", "1", ",", "0"}], "]"}], ",", 
                    RowBox[{"Line", "[", 
                      RowBox[{"{", "\[IndentingNewLine]", 
                        RowBox[{
                          RowBox[{"{", 
                            RowBox[{
                              RowBox[{"5", 
                                RowBox[{"Cos", "[", "tt", "]"}]}], ",", 
                              RowBox[{"5", 
                                RowBox[{"Sin", "[", "tt", "]"}]}]}], "}"}], 
                          ",", 
                          RowBox[{"{", 
                            RowBox[{
                              RowBox[{"5", 
                                RowBox[{"Cos", "[", 
                                  RowBox[{"tt", "+", "\[Pi]"}], "]"}]}], ",", 
                              
                              RowBox[{"5", 
                                RowBox[{"Sin", "[", 
                                  RowBox[{"tt", "+", "\[Pi]"}], "]"}]}]}], 
                            "}"}]}], "}"}], "]"}], ",", "\[IndentingNewLine]",
                     
                    RowBox[{"RGBColor", "[", 
                      RowBox[{"0", ",", "0", ",", "1"}], "]"}], ",", 
                    RowBox[{"Line", "[", 
                      RowBox[{"{", 
                        RowBox[{
                          RowBox[{"{", 
                            RowBox[{
                              RowBox[{"5", 
                                RowBox[{"Cos", "[", 
                                  RowBox[{"tt", "+", 
                                    RowBox[{"\[Pi]", "/", "2"}]}], "]"}]}], 
                              ",", 
                              RowBox[{"5", 
                                RowBox[{"Sin", "[", 
                                  RowBox[{"tt", "+", 
                                    RowBox[{"\[Pi]", "/", "2"}]}], "]"}]}]}], 
                            "}"}], ",", 
                          RowBox[{"{", 
                            RowBox[{
                              RowBox[{"5", 
                                RowBox[{"Cos", "[", 
                                  RowBox[{"tt", "+", 
                                    RowBox[{"3", 
                                      RowBox[{"\[Pi]", "/", "2"}]}]}], 
                                  "]"}]}], ",", 
                              RowBox[{"5", 
                                RowBox[{"Sin", "[", 
                                  RowBox[{"tt", "+", 
                                    RowBox[{"3", 
                                      RowBox[{"\[Pi]", "/", "2"}]}]}], 
                                  "]"}]}]}], "}"}]}], "}"}], "]"}], ",", 
                    "\[IndentingNewLine]", 
                    RowBox[{"RGBColor", "[", 
                      RowBox[{"1", ",", "0", ",", "0"}], "]"}], ",", 
                    "\[IndentingNewLine]", 
                    RowBox[{"Disk", "[", 
                      RowBox[{
                        RowBox[{"{", 
                          RowBox[{
                            RowBox[{"5", 
                              RowBox[{"Cos", "[", "tt", "]"}]}], ",", 
                            RowBox[{"5", 
                              RowBox[{"Sin", "[", "tt", "]"}]}]}], "}"}], ",",
                         "0.4"}], "]"}], ",", "\[IndentingNewLine]", 
                    RowBox[{"Disk", "[", 
                      RowBox[{
                        RowBox[{"{", 
                          RowBox[{
                            RowBox[{"5", 
                              RowBox[{"Cos", "[", 
                                RowBox[{"tt", "+", 
                                  RowBox[{"\[Pi]", "/", "2"}]}], "]"}]}], ",",
                             
                            RowBox[{"5", 
                              RowBox[{"Sin", "[", 
                                RowBox[{"tt", "+", 
                                  RowBox[{"\[Pi]", "/", "2"}]}], "]"}]}]}], 
                          "}"}], ",", "0.4"}], "]"}], ",", 
                    "\[IndentingNewLine]", 
                    RowBox[{"Disk", "[", 
                      RowBox[{
                        RowBox[{"{", 
                          RowBox[{
                            RowBox[{"5", 
                              RowBox[{"Cos", "[", 
                                RowBox[{"tt", "+", "\[Pi]"}], "]"}]}], ",", 
                            RowBox[{"5", 
                              RowBox[{"Sin", "[", 
                                RowBox[{"tt", "+", "\[Pi]"}], "]"}]}]}], 
                          "}"}], ",", "0.4"}], "]"}], ",", 
                    "\[IndentingNewLine]", 
                    RowBox[{"Disk", "[", 
                      RowBox[{
                        RowBox[{"{", 
                          RowBox[{
                            RowBox[{"5", 
                              RowBox[{"Cos", "[", 
                                RowBox[{"tt", "+", 
                                  RowBox[{"3", 
                                    RowBox[{"\[Pi]", "/", "2"}]}]}], "]"}]}], 
                            ",", 
                            RowBox[{"5", 
                              RowBox[{"Sin", "[", 
                                RowBox[{"tt", "+", 
                                  RowBox[{"3", 
                                    RowBox[{"\[Pi]", "/", "2"}]}]}], 
                                "]"}]}]}], "}"}], ",", "0.4"}], "]"}]}], 
                  "\[IndentingNewLine]", "}"}], ",", "\[IndentingNewLine]", 
                RowBox[{"AspectRatio", "\[Rule]", "1"}], ",", 
                RowBox[{"PlotRange", "\[Rule]", 
                  RowBox[{"{", 
                    RowBox[{
                      RowBox[{"{", 
                        RowBox[{
                          RowBox[{"-", "10"}], ",", "10"}], "}"}], ",", 
                      RowBox[{"{", 
                        RowBox[{
                          RowBox[{"-", "10"}], ",", "10"}], "}"}]}], 
                    "}"}]}]}], "]"}]}], "\[IndentingNewLine]", "]"}]}], 
      ";"}], "\[IndentingNewLine]", 
    RowBox[{
      RowBox[{
        RowBox[{"myAnimation", "[", "]"}], ":=", 
        RowBox[{"Update\[Breve]Graphics", "[", 
          RowBox[{"myPlot", "[", "]"}], "]"}]}], ";"}]}], "Input",
  CellLabel->"In[6]:=",
  CellTags->"IGP_7a"],

Cell[BoxData[
    RowBox[{"Interactive\[Breve]Graphics", "[", 
      RowBox[{
        RowBox[{"myPlot", "[", "]"}], ",", 
        RowBox[{"HideNotebook", "\[Rule]", "True"}], ",", 
        RowBox[{"AnimationFunction", "\[Rule]", "myAnimation"}]}], 
      "]"}]], "Input",
  CellLabel->"In[9]:=",
  CellTags->"IGP_7a"],

Cell[TextData[{
  "Because the ",
  StyleBox["HideNotebook",
    FontWeight->"Bold"],
  " option has been set to ",
  StyleBox["True",
    FontWeight->"Bold"],
  ", the notebook disappears for the duration of the display."
}], "Text",
  CellTags->"IGP_7a"],

Cell[TextData[{
  "Note that the ",
  StyleBox["myPlot",
    FontWeight->"Bold"],
  " function will be used before the IGP has even been entered (and therefore \
no animation is started), however, since the ",
  StyleBox["Animation\[Breve]Time",
    FontWeight->"Bold"],
  " function returns zero in this situation, no special coding is needed to \
cover this situation."
}], "Text",
  CellTags->"IGP_7a"],

Cell["\<\
Observe that you can stretch (or even maximise) the window containing the \
rotating graphic, but, because the entire window is refreshed for each frame, \
the animation is less effective at large sizes. Even though the animation \
becomes more jerky, the rate at which the wheel rotates will be unchanged \
\[Dash] it will simply be rendered with fewer frames.\
\>", "Text",
  CellTags->"IGP_7a"],

Cell[BoxData[GridBox[{
        {
          StyleBox["Function",
            "SBO"], 
          StyleBox[
            RowBox[{" ", "Meaning"}],
            "SBO"]},
        {
          RowBox[{"Animation\[Breve]Time", 
            StyleBox["[",
              "MR"], 
            StyleBox["]",
              "MR"]}], 
          RowBox[{
            RowBox[{
            "Returns", " ", "the", " ", "time", " ", "in", " ", "seconds", 
              " ", "since", " ", "the", " ", "start", " ", "of", " ", "the", 
              " ", "animation"}], ",", " ", "\[IndentingNewLine]", 
            RowBox[{
            "or", " ", "zero", " ", "if", " ", "no", " ", "animation", " ", 
              "is", " ", "currently", " ", 
              RowBox[{"running", "."}]}]}]}
        },
      ColumnAlignments->{Left}]], "Text",
  Background->RGBColor[1, 0.8, 0.8],
  CellTags->"IGP_7a"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Using a timer", "Subsubsection",
  CellTags->"IGP_8"],

Cell[TextData[{
  "An animation that is a smooth function of time is best handled using the \
",
  StyleBox["AnimationFunction",
    FontWeight->"Bold"],
  " option as this will ensure that the maximum number of frames are \
generated \[Dash] making the animation as smooth as possible. However, there \
is another type of animation exemplified by a program which reads data off \
the internet to update a graph. Here you probably do not want to update the \
graph more frequently than once a minute (say). The answer is to supply a \
timer function with the ",
  StyleBox["TimerFunction",
    FontWeight->"Bold"],
  " option and possibly a timer interval (in seconds) with the ",
  StyleBox["TimerInterval",
    FontWeight->"Bold"],
  " option. The function that gets called in this way takes no arguments and \
may, or may not, call ",
  StyleBox["Update\[Breve]Graphics",
    FontWeight->"Bold"],
  " to change the displayed image. Here is the outline (i.e. non-functional) \
of an example:"
}], "Text",
  CellTags->"IGP_8"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        RowBox[{"updateData", "[", "]"}], ":=", 
        RowBox[{"Update\[Breve]Graphics", "[", 
          RowBox[{"makeSomePlotFromInternet", "[", "]"}], "]"}]}], 
      ";"}]], "Input",
  CellTags->"IGP_8"],

Cell[BoxData[
    RowBox[{"Interactive\[Breve]Graphics", "[", 
      RowBox[{
        RowBox[{"myPlot", "[", "]"}], ",", 
        RowBox[{"HideNotebook", "\[Rule]", "True"}], ",", 
        RowBox[{"TimerFunction", "\[Rule]", "updateData"}], ",", 
        RowBox[{"TimerInterval", "\[Rule]", "60"}]}], "]"}]], "Input",
  CellTags->"IGP_8"],

Cell[TextData[{
  "\[FilledSmallCircle]\tIt is not necessary that the timer function (or \
indeed the animation function) call ",
  StyleBox["Update\[Breve]Graphics ",
    FontWeight->"Bold"],
  "\[Dash]",
  StyleBox[" ",
    FontWeight->"Bold"],
  "if, for example, no change in the image was required in certain \
circumstances."
}], "Text",
  CellTags->"IGP_8"],

Cell["\<\
\[FilledSmallCircle]\tThe timer interval should be thought of as a lower \
bound, in the sense that the IGP will not call the timer function again until \
the previous call has completed.\
\>", "Text",
  CellTags->"IGP_8"],

Cell[TextData[{
  "\[FilledSmallCircle]\t",
  "It is possible to use both the ",
  StyleBox["AnimationFunction",
    FontWeight->"Bold"],
  " and ",
  StyleBox["TimerFunction",
    FontWeight->"Bold"],
  " options. In this case the ",
  StyleBox["Animation\[Breve]Time",
    FontWeight->"Bold"],
  " function will always return the time since the original animation began \
\[Dash] even if it was replaced by the subsequent operation of a timer \
function."
}], "Text",
  CellTags->"IGP_8"],

Cell["\<\
\[FilledSmallCircle]\tThere is no concept of multi-tasking here \[Dash] each \
function called by the IGP completes its action before another function is \
invoked.\
\>", "Text",
  CellTags->"IGP_8"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Performance considerations", "Subsubsection",
  CellTags->"IGP_9"],

Cell["\<\
When the IGP runs in animation mode it continually refreshes the display \
using all available CPU power. This can, of course, starve other applications \
of CPU resource, and \[Dash] more significantly \[Dash] will drain a portable \
battery quite rapidly and cause the CPU fan to run on most machines. This is \
not a problem unless an animation is left executing for an extended period of \
time.\
\>", "Text",
  CellTags->"IGP_9"],

Cell[TextData[{
  "Where an animated graphic is not expected to update rapidly, it can be \
helpful to use the ",
  StyleBox["Pause",
    FontWeight->"Bold"],
  " function within the animation update function to avoid excessive \
computation."
}], "Text",
  CellTags->"IGP_9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The need for calibration", "Section",
  CellTags->"IGP_10"],

Cell[TextData[{
  "Ordinary ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " graphics is quite unlike most computer graphics in that you can create a \
plot or a low-level graphical object without any thought about the scale \
necessary to place it on the screen or paper. The system chooses a scale for \
you just before the image is created. You can create a plot of part of the \
Milky Way (or part of a molecule) using units of meters if you wish \[Dash] \
the system will simply take care of the necessary scaling. The IGP determines \
what this scale is by displaying a slightly modified version of your plot \
with certain colours adjusted and two small spots added. The resultant image \
is then read back into ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " as an array of pixels, and used to determine the relationship between \
mouse position and graphics coordinates. After this calibration has been \
performed the original image is displayed, and on a reasonably fast machine \
you have to watch quite carefully to notice what is going on. Typically, this \
calibration is only done once, and the IGP 'assumes' that subsequent images \
will require the same calibration. It is important that your graphics satisfy \
this assumption, otherwise the mouse coordinates returned by the system will \
be wrong. For example, if you initially display Plot[Sin[x],{x,0,10}] and \
subsequently use Update\[Breve]Graphics to display Plot[x^2,{x,0,10}], it is \
vital to use the option ReCalibrate->True on the call to Update",
  "\[Breve]",
  "Graphics. Furthermore, if you create animated graphics using IGPTime, it \
is vital that the each frame of the animation uses the same scale. If \
necessary it may be useful to bound your graphics with a box (and not draw \
outside the box!) to ensure that this is so. If your animation seems to jump \
around in the frame, you have a problem here."
}], "Text",
  CellTags->"IGP_10"],

Cell[BoxData[GridBox[{
        {
          StyleBox[
            RowBox[{"Option", " ", "to", " ", "Update\[Breve]Graphics"}],
            "SBO"], 
          StyleBox[
            RowBox[{" ", "Meaning"}],
            "SBO"], 
          StyleBox[
            RowBox[{"Default", " ", "value"}],
            "SBO"]},
        {
          StyleBox["ReCalibrate",
            "MR"], 
          RowBox[{
            RowBox[{
            "Set", " ", "true", " ", "to", " ", "force", " ", "a", " ", 
              "recalibration"}], " ", "\[IndentingNewLine]", 
            RowBox[{"of", " ", "the", " ", 
              RowBox[{"plot", "."}]}]}], "False"}
        },
      ColumnAlignments->{Left}]], "Text",
  Background->RGBColor[1, 0.8, 0.8],
  CellTags->"IGP_10"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Adding a menu", "Section",
  CellTags->"IGP_10X"],

Cell[TextData[{
  "By default, an IGP window has no menu. However, by adding a menu you can \
greatly enrich the amount of control available. The ",
  StyleBox["MenuBar",
    FontWeight->"Bold"],
  " option supplies a menu to the ",
  StyleBox["InteractiveGraphics",
    FontWeight->"Bold"],
  " function. A menu consists of a list of menu items, where a menu item can \
be one of the following:\n\[FilledSmallCircle]\t{<name>,<function>}\n\
\[FilledSmallCircle]\t{<name>,<menu>}\n\[FilledSmallCircle]\tSeparator\n\nThe \
name should be supplied as a string, and the function should be a function \
with zero arguments which is called each time the user selects that menu \
item.\nNotice that the above definition is recursive, in that menus can \
contain sub-menus. This is normal, since menus are typically nested at least \
once. For example:"
}], "Text",
  CellTags->"IGP_10X"],

Cell[BoxData[
    RowBox[{
      RowBox[{"myMenu", "=", 
        RowBox[{"{", 
          RowBox[{"{", 
            RowBox[{"\"\<File\>\"", ",", 
              RowBox[{"{", 
                RowBox[{
                  RowBox[{"{", 
                    RowBox[{"\"\<Open\>\"", ",", "open\[Breve]func"}], "}"}], 
                  ",", 
                  RowBox[{"{", 
                    RowBox[{"\"\<Save\>\"", ",", "save_func"}], "}"}], ",", 
                  RowBox[{"{", 
                    RowBox[{"\"\<Exit\>\"", ",", "exit\[Breve]func"}], 
                    "}"}]}], "}"}]}], "}"}], "}"}]}], ";"}]], "Input",
  CellTags->"IGP_10X"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Hiding the notebook", "Section",
  CellTags->"IGP_11"],

Cell[TextData[{
  "The IGP operates using J/Link, which creates Java windows as a separate \
process. This works well except that there is no way to prevent the IGP \
window being covered over by the notebook. Typically this will happen when \
the user clicks on his notebook. This can be quite troublesome if the user \
does not realise what has happened, as ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " is simply servicing the interactive graphics, but this is overlaid by the \
notebook! The option ",
  StyleBox["HideNotebook",
    FontWeight->"Bold"],
  " can be used to make the notebook that started the calculation (the \
EvaluationNotebook[]) invisible for the duration of the graphical \
interaction. This can be visually extremely effective, but observe the \
following caveats:"
}], "Text",
  CellTags->"IGP_11"],

Cell[TextData[{
  "\[FilledSmallCircle]\tOnly use this option after a program has been \
reasonably well tested. If your code aborts or gets stuck it is possible that \
your notebook will remain hidden even though the graphics window has closed. \
This could easily prevent you saving any changes in the notebook! If this \
happens, you can recover by opening a temporary notebook, finding the \
notebook in question using Notebooks[], and then manually setting the option \
Visible->True on this notebook. Clearly prevention is easier than cure!\n\
\[FilledSmallCircle]\tIf you shut down ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " with a non-visible notebook open, the system will not warn you if it has \
been modified but not saved.\n\[FilledSmallCircle]\tThe ",
  StyleBox["HideNotebook",
    FontWeight->"Bold"],
  " option only hides the notebook that began the evaluation. If you want to \
hide all notebooks, you can do this by setting the option Visible->False to \
every notebook in the list returned by Notebooks[]. \n\[FilledSmallCircle]\t\
Although the ",
  StyleBox["HideNotebook",
    FontWeight->"Bold"],
  " option has only been used on certain examples in this guide, it could be \
usefully used with any of the other examples."
}], "Text",
  CellTags->"IGP_11"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Other functions", "Section",
  CellTags->"IGP_12"],

Cell["\<\
The following functions can be called while a graphic is being displayed \
(typically in response to a mouse movement).\
\>", "Text",
  CellTags->"IGP_12"],

Cell[BoxData[GridBox[{
        {
          StyleBox["Function",
            "SBO"], 
          StyleBox[
            RowBox[{" ", "Meaning"}],
            "SBO"]},
        {
          StyleBox[
            RowBox[{"Close\[Breve]Graphics\[Breve]Window", "[", "]"}],
            "MR"], 
          RowBox[{
            RowBox[{
            "Closes", " ", "the", " ", "graphics", " ", "window", " ", "as", 
              " ", "if", " ", "the", " ", "close", " ", "button", " ", 
              "had"}], "\[IndentingNewLine]", 
            RowBox[{
            "been", " ", "pressed", " ", "\[Dash]", " ", "typically", " ", 
              "called", " ", "in", " ", "response", " ", "to", " ", 
              RowBox[{"an", " ", "'"}], 
              RowBox[{"Exit", "'"}]}], " ", "\[IndentingNewLine]", 
            RowBox[{"menu", " ", 
              RowBox[{"item", "."}]}]}]},
        {
          StyleBox[
            RowBox[{
            "Get\[Breve]Mouse\[Breve]Click\[Breve]Identifier", "[", "]"}],
            "MR"], 
          RowBox[{
            RowBox[{
            "Returns", " ", "an", " ", "integer", " ", "identification", " ", 
              "code", " ", "which", " ", "distinguishes"}], 
            "\[IndentingNewLine]", 
            RowBox[{
            "between", " ", "the", " ", "various", " ", "mouse", " ", 
              "buttons", " ", "and", " ", "whether", " ", "any", " ", "of", 
              " ", "the", " ", "various"}], "\[IndentingNewLine]", 
            RowBox[{
            "keyboard", " ", "shift", " ", "keys", " ", "have", " ", "been", 
              " ", 
              RowBox[{"pressed", "."}]}]}]},
        {
          StyleBox[
            RowBox[{"Set\[Breve]Default\[Breve]Cursor", "[", "]"}],
            "MR"], 
          RowBox[{
          "Sets", " ", "the", " ", "graphics", " ", "cursor", " ", "to", " ", 
            "the", " ", "default", " ", 
            RowBox[{"pointer", "."}]}]},
        {
          StyleBox[
            RowBox[{"Set\[Breve]Hand\[Breve]Cursor", "[", "]"}],
            "MR"], 
          RowBox[{
            RowBox[{
              RowBox[{
              "Sets", " ", "the", " ", "graphics", " ", "cursor", " ", "to", 
                " ", "a", " ", "hand", " ", "symbol"}], " ", "-", " ", 
              RowBox[{"typically", " ", "used"}]}], " ", 
            "\[IndentingNewLine]", 
            RowBox[{"to", " ", "denote", " ", "a", " ", "clickable", " ", 
              RowBox[{"button", "."}]}]}]},
        {
          StyleBox[
            RowBox[{"Set\[Breve]Wait\[Breve]Cursor", "[", "]"}],
            "MR"], 
          RowBox[{
            RowBox[{
              RowBox[{
              "Sets", " ", "the", " ", "graphics", " ", "cursor", " ", "to", 
                " ", "the", " ", "hour", " ", "glass", " ", "symbol"}], " ", 
              "-", " ", "typically"}], " ", "\[IndentingNewLine]", 
            RowBox[{
            "used", " ", "to", " ", "denote", " ", "an", " ", "operation", 
              " ", "that", " ", "is", " ", "not", " ", "yet", " ", 
              RowBox[{"complete", "."}]}]}]}
        },
      ColumnAlignments->{Left}]], "Text",
  Background->RGBColor[1, 0.8, 0.8],
  CellTags->"IGP_12"],

Cell[TextData[{
  "The ",
  StyleBox["Get\[Breve]Mouse\[Breve]Click\[Breve]Identifier ",
    FontWeight->"Bold"],
  "function should only be called while processing a mouse click. It is \
designed for fairly complex programs in which you want to respond to a mouse \
click in several different ways depending on which mouse button was pressed \
and which 'shift' keyboard keys such as \[ControlKey], \[AltKey], \[ShiftKey] \
were depressed at the time of the click. Be aware that the number of mouse \
buttons and types of keyboard keys are both platform dependant."
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Some larger examples", "Section",
  CellTags->"IGP_13"],

Cell[CellGroupData[{

Cell["Adding and deleting data from a ListPlot", "Subsubsection",
  CellTags->"IGP_14"],

Cell["\<\
Here we see an extension of the above example. We now have too modes \[Dash] \
selectable by the large red button \[Dash] so that it is possible both add \
and delete points. When two or more points are displayed, the software plots \
a best-fit straight line through the data. You can watch this move about as \
you add or remove data.
The function overButton is used to determine if the mouse is positioned over \
the button or over the graph itself. A mouse-motion handler has been attached \
to change the cursor shape over the button. Try moving the mouse cursor over \
the button to observe this effect. The button has been included mainly for \
illustration, as the menu can also be used to select the mode.
Notice how the title of the plot is made to depend on the variable \
addingPoints, which is toggled each time the button is pressed. \
\>", "Text",
  CellTags->"IGP_14"],

Cell[BoxData[
    RowBox[{"Needs", "[", "\"\<InteractiveGraphics`\>\"", "]"}]], "Input",
  CellLabel->"In[1]:=",
  CellTags->"IGP_6"],

Cell[BoxData[{
    StyleBox[
      RowBox[{
        RowBox[{"Clear", "[", 
          RowBox[{"myPlot", ",", "mouseUpdate", ",", "myClickCriterion"}], 
          "]"}], ";"}],
      FontColor->GrayLevel[0]], "\n", 
    RowBox[{
      RowBox[{
        StyleBox["buttonCoords",
          FontColor->GrayLevel[0]], 
        StyleBox["=",
          ShowContents->True], 
        StyleBox[
          RowBox[{"{", 
            RowBox[{
              RowBox[{"{", 
                RowBox[{"1", ",", 
                  RowBox[{"-", "4"}]}], "}"}], ",", 
              RowBox[{"{", 
                RowBox[{"3", ",", 
                  RowBox[{"-", "2"}]}], "}"}]}], "}"}],
          FontColor->GrayLevel[0]]}], 
      StyleBox[";",
        FontColor->GrayLevel[0]]}], "\n", 
    RowBox[{
      RowBox[{
        StyleBox["addingPoints",
          FontColor->GrayLevel[0]], 
        StyleBox["=",
          ShowContents->True,
          FontColor->GrayLevel[0]], 
        StyleBox["True",
          FontColor->GrayLevel[0]]}], 
      StyleBox[";",
        FontColor->GrayLevel[0]]}], "\n", 
    RowBox[{
      RowBox[{
        RowBox[{
          StyleBox["myData",
            FontColor->GrayLevel[0]], 
          StyleBox["=",
            ShowContents->True], 
          StyleBox[
            RowBox[{"{", 
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"1", ",", "1"}], "}"}], ",", 
                RowBox[{"{", 
                  RowBox[{"4", ",", "4"}], "}"}]}], "}"}],
            FontColor->GrayLevel[0]]}], 
        StyleBox[";",
          FontColor->GrayLevel[0]]}], 
      StyleBox["\n",
        FontColor->GrayLevel[0]]}], "\n", 
    StyleBox[
      RowBox[{
        RowBox[{
          RowBox[{
            RowBox[{"myClickCriterion", "[", 
              RowBox[{
                RowBox[{"{", 
                  RowBox[{"a_", ",", "b_"}], "}"}], ",", 
                RowBox[{"{", 
                  RowBox[{"c_", ",", "d_"}], 
                  StyleBox["}",
                    FontColor->GrayLevel[0]]}]}], 
              StyleBox["]",
                FontColor->GrayLevel[0]]}], 
            StyleBox[":=",
              ShowContents->True], 
            RowBox[{
              StyleBox["(",
                FontColor->GrayLevel[0]], 
              RowBox[{
                RowBox[{
                  RowBox[{
                    RowBox[{
                      StyleBox["(",
                        FontColor->GrayLevel[0]], 
                      RowBox[{
                        StyleBox["a",
                          FontColor->GrayLevel[0]], 
                        StyleBox[" ",
                          FontColor->GrayLevel[0]], 
                        StyleBox["-",
                          FontColor->GrayLevel[0]], " ", "c"}], ")"}], "^", 
                    "2"}], " ", "+", " ", 
                  RowBox[{
                    RowBox[{"(", 
                      RowBox[{"b", " ", "-", " ", "d"}], ")"}], "^", "2"}]}], 
                "<", " ", "0.1"}], ")"}]}], ";"}], "\[IndentingNewLine]"}],
      FontColor->GrayLevel[0]], "\n", 
    RowBox[{
      RowBox[{
        RowBox[{
          StyleBox[
            RowBox[{"myPlot", "[", "]"}],
            FontColor->GrayLevel[0]], 
          StyleBox[":=",
            ShowContents->True], 
          RowBox[{
            StyleBox["Module",
              ShowContents->True], 
            StyleBox["[",
              FontColor->GrayLevel[0]], 
            RowBox[{
              RowBox[{
                StyleBox["{",
                  FontColor->GrayLevel[0]], 
                RowBox[{
                  StyleBox["m",
                    FontColor->GrayLevel[0]], 
                  StyleBox[",",
                    FontColor->GrayLevel[0]], 
                  StyleBox["c",
                    FontColor->GrayLevel[0]], 
                  StyleBox[",",
                    FontColor->GrayLevel[0]], 
                  RowBox[{
                    StyleBox["fitting",
                      FontColor->GrayLevel[0]], 
                    StyleBox["=",
                      ShowContents->True,
                      FontColor->GrayLevel[0]], 
                    StyleBox["False",
                      FontColor->GrayLevel[0]]}], 
                  StyleBox[",",
                    FontColor->GrayLevel[0]], 
                  StyleBox["tmp",
                    FontColor->GrayLevel[0]], 
                  StyleBox[",",
                    FontColor->GrayLevel[0]], 
                  StyleBox["p1",
                    FontColor->GrayLevel[0]], 
                  StyleBox[",",
                    FontColor->GrayLevel[0]], 
                  StyleBox["p2",
                    FontColor->GrayLevel[0]]}], 
                StyleBox["}",
                  FontColor->GrayLevel[0]]}], 
              StyleBox[",",
                FontColor->GrayLevel[0]], 
              StyleBox["\[IndentingNewLine]",
                FontColor->GrayLevel[0]], 
              RowBox[{
                RowBox[{
                  StyleBox["If",
                    ShowContents->True], 
                  StyleBox["[",
                    FontColor->GrayLevel[0]], 
                  RowBox[{
                    StyleBox[
                      RowBox[{
                        RowBox[{"Length", "[", "myData", "]"}], ">", "1"}],
                      FontColor->GrayLevel[0]], 
                    StyleBox[",",
                      FontColor->GrayLevel[0]], 
                    StyleBox["\[IndentingNewLine]",
                      FontColor->GrayLevel[0]], 
                    RowBox[{
                      RowBox[{
                        StyleBox["fitting",
                          FontColor->GrayLevel[0]], 
                        StyleBox["=",
                          ShowContents->True,
                          FontColor->GrayLevel[0]], 
                        StyleBox["True",
                          FontColor->GrayLevel[0]]}], 
                      StyleBox[";",
                        FontColor->GrayLevel[0]], 
                      StyleBox["\[IndentingNewLine]",
                        FontColor->GrayLevel[0]], 
                      RowBox[{
                        StyleBox["tmp",
                          FontColor->GrayLevel[0]], 
                        StyleBox["=",
                          ShowContents->True], 
                        StyleBox[
                          RowBox[{"FindFit", "[", 
                            RowBox[{"myData", ",", 
                              RowBox[{"c", "+", 
                                RowBox[{"m", " ", "x"}]}], ",", 
                              RowBox[{"{", 
                                RowBox[{"c", ",", "m"}], "}"}], ",", "x"}], 
                            "]"}],
                          FontColor->GrayLevel[0]]}], 
                      StyleBox[";",
                        FontColor->GrayLevel[0]], 
                      StyleBox["\[IndentingNewLine]",
                        FontColor->GrayLevel[0]], 
                      RowBox[{
                        StyleBox["p1",
                          FontColor->GrayLevel[0]], 
                        StyleBox["=",
                          ShowContents->True], 
                        StyleBox[
                          RowBox[{
                            RowBox[{"{", 
                              RowBox[{"0", ",", "c"}], "}"}], "/.", "tmp"}],
                          FontColor->GrayLevel[0]]}], 
                      StyleBox[";",
                        FontColor->GrayLevel[0]], 
                      StyleBox["\[IndentingNewLine]",
                        FontColor->GrayLevel[0]], 
                      RowBox[{
                        StyleBox["p2",
                          FontColor->GrayLevel[0]], 
                        StyleBox["=",
                          ShowContents->True], 
                        StyleBox[
                          RowBox[{
                            RowBox[{"{", 
                              RowBox[{"10", ",", 
                                RowBox[{
                                  RowBox[{"10", " ", "m"}], "+", "c"}]}], 
                              "}"}], "/.", "tmp"}],
                          FontColor->GrayLevel[0]]}], 
                      StyleBox[";",
                        FontColor->GrayLevel[0]]}]}], 
                  StyleBox["\[IndentingNewLine]",
                    FontColor->GrayLevel[0]], 
                  StyleBox["]",
                    FontColor->GrayLevel[0]]}], 
                StyleBox[";",
                  FontColor->GrayLevel[0]], "\[IndentingNewLine]", 
                RowBox[{"ListPlot", "[", 
                  RowBox[{"myData", ",", 
                    RowBox[{"PlotRange", "\[Rule]", 
                      RowBox[{"{", 
                        RowBox[{
                          RowBox[{"{", 
                            RowBox[{"0", ",", "10"}], "}"}], ",", 
                          RowBox[{"{", 
                            RowBox[{
                              RowBox[{"-", "5"}], ",", "18"}], "}"}]}], 
                        "}"}]}], ",", "\[IndentingNewLine]", 
                    RowBox[{"Epilog", "\[Rule]", 
                      RowBox[{"{", 
                        RowBox[{
                          RowBox[{
                            StyleBox["If",
                              ShowContents->True], "[", 
                            RowBox[{"fitting", ",", 
                              RowBox[{"{", 
                                RowBox[{
                                  RowBox[{"RGBColor", "[", 
                                    RowBox[{"1", ",", "0", ",", "0"}], "]"}], 
                                  ",", 
                                  RowBox[{"Line", "[", 
                                    RowBox[{"{", 
                                      RowBox[{"p1", ",", "p2"}], "}"}], 
                                    "]"}]}], "}"}], ",", 
                              RowBox[{"{", "}"}]}], "]"}], ",", 
                          "\[IndentingNewLine]", 
                          RowBox[{"Text", "[", 
                            RowBox[{
                              RowBox[{
                                StyleBox["If",
                                  ShowContents->True], "[", 
                                RowBox[{
                                "addingPoints", ",", 
                                  "\"\<Use the mouse to add points\>\"", ",", 
                                  "\"\<Use the mouse to remove points\>\""}], 
                                "]"}], ",", 
                              RowBox[{"{", 
                                RowBox[{"5", ",", "17"}], "}"}], ",", 
                              RowBox[{"{", 
                                RowBox[{"0", ",", "0"}], "}"}], ",", 
                              RowBox[{"TextStyle", "\[Rule]", 
                                RowBox[{"{", 
                                  RowBox[{"FontSize", "\[Rule]", "17"}], 
                                  "}"}]}]}], "]"}], ",", 
                          "\[IndentingNewLine]", 
                          RowBox[{"RGBColor", "[", 
                            RowBox[{"1", ",", "0", ",", "0"}], "]"}], ",", 
                          RowBox[{"Rectangle", "@@", "buttonCoords"}], ",", 
                          RowBox[{"RGBColor", "[", 
                            RowBox[{"0", ",", "0", ",", "0"}], "]"}], ",", 
                          "\[IndentingNewLine]", 
                          RowBox[{"Text", "[", 
                            RowBox[{"\"\<Select mode\>\"", ",", 
                              RowBox[{
                                RowBox[{"Plus", "@@", "buttonCoords"}], "/", 
                                "2"}], ",", 
                              RowBox[{"{", 
                                RowBox[{"0", ",", "0"}], "}"}], ",", 
                              RowBox[{"TextStyle", "\[Rule]", 
                                RowBox[{"{", 
                                  RowBox[{"FontSize", "\[Rule]", "14"}], 
                                  "}"}]}]}], "]"}]}], "}"}]}], ",", 
                    "\[IndentingNewLine]", 
                    RowBox[{"DisplayFunction", "\[Rule]", "Identity"}], ",", 
                    RowBox[{"ImageSize", "\[Rule]", 
                      RowBox[{"{", 
                        RowBox[{"600", ",", "600"}], "}"}]}], ",", 
                    RowBox[{"PlotStyle", "\[Rule]", 
                      RowBox[{"{", 
                        RowBox[{"PointSize", "[", "0.01", "]"}], "}"}]}]}], 
                  "]"}]}]}], "\[IndentingNewLine]", "]"}]}], ";"}], 
      "\n"}], "\n", 
    RowBox[{
      RowBox[{
        RowBox[{
          RowBox[{"overButton", "[", 
            RowBox[{"x_", ",", "y_"}], "]"}], 
          StyleBox[":=",
            ShowContents->True], 
          RowBox[{
            RowBox[{"x", "\[GreaterEqual]", 
              RowBox[{"buttonCoords", 
                StyleBox["\[LeftDoubleBracket]",
                  ShowContents->True], 
                RowBox[{"1", ",", "1"}], 
                StyleBox["\[RightDoubleBracket]",
                  ShowContents->True]}]}], "&&", 
            RowBox[{"x", "\[LessEqual]", 
              RowBox[{"buttonCoords", 
                StyleBox["\[LeftDoubleBracket]",
                  ShowContents->True], 
                RowBox[{"2", ",", "1"}], 
                StyleBox["\[RightDoubleBracket]",
                  ShowContents->True]}]}], " ", "&&", " ", 
            RowBox[{"y", "\[GreaterEqual]", 
              RowBox[{"buttonCoords", 
                StyleBox["\[LeftDoubleBracket]",
                  ShowContents->True], 
                RowBox[{"1", ",", "2"}], 
                StyleBox["\[RightDoubleBracket]",
                  ShowContents->True]}]}], "&&", 
            RowBox[{"y", "\[LessEqual]", 
              RowBox[{"buttonCoords", 
                StyleBox["\[LeftDoubleBracket]",
                  ShowContents->True], 
                RowBox[{"2", ",", "2"}], 
                StyleBox["\[RightDoubleBracket]",
                  ShowContents->True]}]}]}]}], ";"}], 
      "\n"}], "\[IndentingNewLine]", 
    RowBox[{
      RowBox[{
        RowBox[{
          RowBox[{"mouseCursorControl", "[", 
            RowBox[{"x_", ",", "y_"}], "]"}], 
          StyleBox[":=",
            ShowContents->True], "\[IndentingNewLine]", 
          RowBox[{
            StyleBox["If",
              ShowContents->True], "[", 
            RowBox[{
              RowBox[{"overButton", "[", 
                RowBox[{"x", ",", "y"}], "]"}], ",", 
              RowBox[{"Set\[Breve]Hand\[Breve]Cursor", "[", "]"}], ",", 
              RowBox[{"Set\[Breve]Default\[Breve]Cursor", "[", "]"}]}], 
            "]"}]}], ";"}], "\[IndentingNewLine]"}], "\n", 
    RowBox[{
      RowBox[{
        RowBox[{"setDeleteMode", "[", "]"}], ":=", 
        RowBox[{"(", 
          RowBox[{
            RowBox[{"addingPoints", "=", "False"}], ";", 
            RowBox[{"Update\[Breve]Graphics", "[", 
              RowBox[{"myPlot", "[", "]"}], "]"}]}], ")"}]}], 
      ";"}], "\[IndentingNewLine]", 
    RowBox[{
      RowBox[{
        RowBox[{"setAddMode", "[", "]"}], ":=", 
        RowBox[{"(", 
          RowBox[{
            RowBox[{"addingPoints", "=", "True"}], ";", 
            RowBox[{"Update\[Breve]Graphics", "[", 
              RowBox[{"myPlot", "[", "]"}], "]"}]}], ")"}]}], 
      ";"}], "\[IndentingNewLine]", 
    RowBox[{
      RowBox[{
        RowBox[{"myExit", "[", "]"}], ":=", 
        RowBox[{"Close\[Breve]Graphics\[Breve]Window", "[", "]"}]}], 
      ";"}], "\[IndentingNewLine]", 
    RowBox[{
      RowBox[{
        RowBox[{"mouseClick", "[", 
          RowBox[{"x_", ",", "y_"}], "]"}], 
        StyleBox[":=",
          ShowContents->True], 
        RowBox[{
          StyleBox["Module",
            ShowContents->True], "[", 
          RowBox[{
            RowBox[{"{", "pt", "}"}], ",", "\[IndentingNewLine]", 
            RowBox[{
              StyleBox["If",
                ShowContents->True], "[", 
              RowBox[{
                RowBox[{"overButton", "[", 
                  RowBox[{"x", ",", "y"}], "]"}], ",", "\[IndentingNewLine]", 
                
                RowBox[{
                  RowBox[{"addingPoints", 
                    StyleBox["=",
                      ShowContents->True], 
                    RowBox[{"(", 
                      RowBox[{"!", "addingPoints"}], ")"}]}], ";", 
                  "\[IndentingNewLine]", 
                  RowBox[{"Update\[Breve]Graphics", "[", 
                    RowBox[{"myPlot", "[", "]"}], "]"}]}], ",", 
                "\[IndentingNewLine]", 
                RowBox[{
                  StyleBox["If",
                    ShowContents->True], "[", 
                  RowBox[{"addingPoints", ",", "\[IndentingNewLine]", 
                    RowBox[{
                      RowBox[{"Print", "[", 
                        RowBox[{"\"\<Adding data point: \>\"", ",", 
                          RowBox[{"{", 
                            RowBox[{"x", ",", "y"}], "}"}]}], "]"}], ";", 
                      "\[IndentingNewLine]", 
                      RowBox[{"myData", 
                        StyleBox["=",
                          ShowContents->True], 
                        RowBox[{"Append", "[", 
                          RowBox[{"myData", ",", 
                            RowBox[{"{", 
                              RowBox[{"x", ",", "y"}], "}"}]}], "]"}]}], ";", 
                      "\[IndentingNewLine]", 
                      RowBox[{"Update\[Breve]Graphics", "[", 
                        RowBox[{"myPlot", "[", "]"}], "]"}]}], ",", 
                    "\[IndentingNewLine]", 
                    RowBox[{
                      RowBox[{"pt", 
                        StyleBox["=",
                          ShowContents->True], 
                        RowBox[{"Select", "[", 
                          RowBox[{"myData", ",", 
                            RowBox[{
                              RowBox[{"myClickCriterion", "[", 
                                RowBox[{
                                  RowBox[{"{", 
                                    RowBox[{"x", ",", "y"}], "}"}], ",", 
                                  "#"}], "]"}], "&"}]}], "]"}]}], ";", 
                      "\[IndentingNewLine]", 
                      RowBox[{
                        StyleBox["If",
                          ShowContents->True], "[", 
                        RowBox[{
                          RowBox[{"pt", " ", "\[NotEqual]", " ", 
                            RowBox[{"{", "}"}]}], ",", "\[IndentingNewLine]", 
                          
                          RowBox[{
                            RowBox[{"pt", 
                              StyleBox["=",
                                ShowContents->True], 
                              RowBox[{"pt", 
                                StyleBox["\[LeftDoubleBracket]",
                                  ShowContents->True], "1", 
                                StyleBox["\[RightDoubleBracket]",
                                  ShowContents->True]}]}], ";", 
                            "\[IndentingNewLine]", 
                            RowBox[{"Print", "[", 
                              RowBox[{
                              "\"\<Deleting data point: \>\"", ",", "pt"}], 
                              "]"}], ";", "\[IndentingNewLine]", 
                            RowBox[{"myData", 
                              StyleBox["=",
                                ShowContents->True], 
                              RowBox[{"DeleteCases", "[", 
                                RowBox[{"myData", ",", "pt"}], "]"}]}], ";", 
                            "\[IndentingNewLine]", 
                            RowBox[{"Update\[Breve]Graphics", "[", 
                              RowBox[{"myPlot", "[", "]"}], "]"}], ";"}]}], 
                        "\[IndentingNewLine]", "]"}]}]}], 
                  "\[IndentingNewLine]", "]"}]}], "\[IndentingNewLine]", 
              "]"}]}], "\[IndentingNewLine]", "]"}]}], 
      ";"}], "\[IndentingNewLine]", 
    RowBox[{
      RowBox[{"myMenu", "=", 
        RowBox[{"{", 
          RowBox[{
            RowBox[{"{", 
              RowBox[{"\"\<File\>\"", ",", 
                RowBox[{"{", 
                  RowBox[{"{", 
                    RowBox[{"\"\<Exit\>\"", ",", "myExit"}], "}"}], "}"}]}], 
              "}"}], ",", 
            RowBox[{"{", 
              RowBox[{"\"\<Options\>\"", ",", 
                RowBox[{"{", 
                  RowBox[{
                    RowBox[{"{", 
                      RowBox[{"\"\<Add mode\>\"", ",", "setAddMode"}], "}"}], 
                    ",", 
                    RowBox[{"{", 
                      RowBox[{"\"\<Delete mode\>\"", ",", "setDeleteMode"}], 
                      "}"}]}], "}"}]}], "}"}]}], "}"}]}], ";"}]}], "Input",
  CellLabel->"In[2]:=",
  CellTags->"IGP_14"],

Cell[CellGroupData[{

Cell[BoxData[
    RowBox[{
      RowBox[{"Interactive\[Breve]Graphics", "[", 
        RowBox[{
          RowBox[{"myPlot", "[", "]"}], ",", 
          RowBox[{"MouseClickFunction", "->", "mouseClick"}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{"MouseMotionFunction", "\[Rule]", "mouseCursorControl"}], 
          ",", "\[IndentingNewLine]", 
          RowBox[{
          "Title", "->", "\"\<Build your own graph with linear fit\>\""}], 
          ",", 
          RowBox[{"MenuBar", "\[Rule]", "myMenu"}]}], "]"}], ";"}]], "Input",
  CellLabel->"In[15]:=",
  CellTags->"IGP_14"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Adding data point: \"\>", "\[InvisibleSpace]", 
        RowBox[{"{", 
          RowBox[{"8.27761614265603`", ",", "8.761126268570903`"}], "}"}]}],
      SequenceForm[ 
      "Adding data point: ", {8.2776161426560293, 8.7611262685709033}],
      Editable->False]], "Print",
  CellLabel->"From In[15]:=",
  CellTags->"IGP_14"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Adding data point: \"\>", "\[InvisibleSpace]", 
        RowBox[{"{", 
          RowBox[{"8.440919755983106`", ",", "7.788634255618057`"}], "}"}]}],
      SequenceForm[ 
      "Adding data point: ", {8.4409197559831064, 7.7886342556180566}],
      Editable->False]], "Print",
  CellLabel->"From In[15]:=",
  CellTags->"IGP_14"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"\<\"Deleting data point: \"\>", "\[InvisibleSpace]", 
        RowBox[{"{", 
          RowBox[{"8.27761614265603`", ",", "8.761126268570903`"}], "}"}]}],
      SequenceForm[ 
      "Deleting data point: ", {8.2776161426560293, 8.7611262685709033}],
      Editable->False]], "Print",
  CellLabel->"From In[15]:=",
  CellTags->"IGP_14"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A naive line drawing program", "Subsubsection",
  CellTags->"IGP_15"],

Cell["\<\
The following illustrates an extremely simple line drawing program. Although \
not very useful in itself, it could form the basis for a program that would \
interactively 'wire up' electronic components, or create a graph (in the \
graph-theoretic sense) of some sort. Remember that there is no reason why the \
lines need be drawn exactly where the mouse travelled, for example, they \
could be made to snap to particular locations etc.\
\>", "Text",
  CellTags->"IGP_15"],

Cell[BoxData[
    RowBox[{"Needs", "[", "\"\<InteractiveGraphics`\>\"", "]"}]], "Input",
  CellLabel->"In[1]:=",
  CellTags->"IGP_6"],

Cell[BoxData[{
    RowBox[{
      RowBox[{"lineList", "=", 
        RowBox[{"{", "}"}]}], ";"}], "\[IndentingNewLine]", 
    RowBox[{
      RowBox[{
        RowBox[{
          RowBox[{"myPlot", "[", "]"}], ":=", 
          RowBox[{"Graphics", "[", 
            RowBox[{
              RowBox[{"Map", "[", 
                RowBox[{"Line", ",", "lineList"}], "]"}], ",", 
              RowBox[{"PlotRange", "\[Rule]", 
                RowBox[{"{", 
                  RowBox[{
                    RowBox[{"{", 
                      RowBox[{"0", ",", "10"}], "}"}], ",", 
                    RowBox[{"{", 
                      RowBox[{"0", ",", "10"}], "}"}]}], "}"}]}], ",", 
              RowBox[{"AspectRatio", "\[Rule]", "1"}]}], "]"}]}], ";"}], 
      "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
    RowBox[{
      RowBox[{
        RowBox[{"myMouseSelector", "[", 
          RowBox[{"x1_", ",", "y1_", ",", "x2_", ",", "y2_"}], "]"}], ":=", 
        RowBox[{"Module", "[", 
          RowBox[{
            RowBox[{"{", "}"}], ",", "\[IndentingNewLine]", 
            RowBox[{
              RowBox[{"lineList", "=", 
                RowBox[{"Append", "[", 
                  RowBox[{"lineList", ",", 
                    RowBox[{"{", 
                      RowBox[{
                        RowBox[{"{", 
                          RowBox[{"x1", ",", "y1"}], "}"}], ",", 
                        RowBox[{"{", 
                          RowBox[{"x2", ",", "y2"}], "}"}]}], "}"}]}], 
                  "]"}]}], ";", "\[IndentingNewLine]", 
              RowBox[{"Update\[Breve]Graphics", "[", 
                RowBox[{"myPlot", "[", "]"}], "]"}], ";"}]}], 
          "\[IndentingNewLine]", "]"}]}], ";"}]}], "Input",
  CellLabel->"In[16]:=",
  CellTags->"IGP_15"],

Cell[BoxData[
    RowBox[{"Interactive\[Breve]Graphics", "[", 
      RowBox[{
        RowBox[{"myPlot", "[", "]"}], ",", 
        RowBox[{"MouseLineSelectionFunction", "\[Rule]", "myMouseSelector"}], 
        ",", "\[IndentingNewLine]", 
        RowBox[{"Title", "->", "\"\<Drag the mouse to draw lines\>\""}], ",", 
        
        RowBox[{"HideNotebook", "\[Rule]", "True"}]}], "]"}]], "Input",
  CellLabel->"In[19]:=",
  CellTags->"IGP_15"],

Cell[TextData[{
  "Notice how when you drag the mouse a temporary line is drawn to guide the \
user. This line is drawn by the IGP because the ",
  StyleBox["MouseLineSelectionFunction",
    FontWeight->"Bold"],
  " option was used. If the ",
  StyleBox["MouseBoxSelectionFunction",
    FontWeight->"Bold"],
  " option had been used instead, the code would work exactly the same except \
that a temporary rectangle would be drawn as the mouse was dragged. This is \
particularly useful for interactively zooming into part of a plot."
}], "Text",
  CellTags->"IGP_15"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Future developments", "Section",
  CellTags->"IGP_16"],

Cell["\<\
Clearly, this simple package can be extended in many ways, and a future \
version may contain some of the following, depending on demand:\
\>", "Text",
  CellTags->"IGP_16"],

Cell["\<\
\[FilledSmallCircle]\tSupport for some Java widgets (e.g. buttons and \
sliders) to control the graphics
\[FilledSmallCircle]\tIntegration with the GUIKit\
\>", "Text",
  CellTags->"IGP_16"],

Cell["\<\
\[FilledSmallCircle]\tSupport for more elaborate animation\
\>", "Text",
  CellTags->"IGP_16"],

Cell["\<\
The IGP is not a large package, and it may be that others will wish to extend \
it. If you do this, please keep the copyright notice intact and add a comment \
to indicate that the package has been modified. The author would be glad to \
incorporate enhancements of general interest in future versions.\
\>", "Text",
  CellTags->"IGP_16"]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"5.0 for Microsoft Windows",
ScreenRectangle->{{0, 1280}, {0, 971}},
Selectable->True,
WindowSize->{1082, 764},
WindowMargins->{{81, Automatic}, {Automatic, -16}},
Visible->True,
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
PageFooters->{{Inherited, Cell[ 
        TextData[ {"\[Copyright]1998 Salford Software Ltd."}], "PageFooter"], 
      Inherited}, {Inherited, Cell[ 
        TextData[ {"\[Copyright]1998 Salford Software Ltd."}], "PageFooter"], 
      Inherited}},
PrivateNotebookOptions->{"ShortBoxForm"->False},
ShowCellBracket->True,
ShowSelection->True,
ShowCellLabel->False,
Magnification->1.5,
StyleDefinitions -> "HelpBrowser.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "IGP_1"->{
    Cell[1904, 58, 88, 2, 82, "Section",
      CellTags->"IGP_1"],
    Cell[2017, 64, 70, 1, 58, "Subsection",
      CellTags->"IGP_1"],
    Cell[2090, 67, 3207, 62, 454, "Text",
      CellTags->"IGP_1"]},
  "IGP_3"->{
    Cell[5434, 139, 68, 1, 82, "Section",
      CellTags->"IGP_3"]},
  "IGP_4"->{
    Cell[5527, 144, 68, 1, 50, "Subsubsection",
      CellTags->"IGP_4"],
    Cell[5598, 147, 985, 21, 152, "Text",
      CellTags->"IGP_4"]},
  "IGP_5"->{
    Cell[6620, 173, 65, 1, 50, "Subsubsection",
      CellTags->"IGP_5"],
    Cell[6688, 176, 89, 1, 34, "Text",
      CellTags->"IGP_5"],
    Cell[6780, 179, 120, 3, 40, "Input",
      CellTags->"IGP_5"]},
  "IGP_6"->{
    Cell[6937, 187, 56, 1, 50, "Subsubsection",
      CellTags->"IGP_6"],
    Cell[6996, 190, 72, 1, 34, "Text",
      CellTags->"IGP_6"],
    Cell[7071, 193, 133, 3, 40, "Input",
      CellTags->"IGP_6"],
    Cell[7207, 198, 615, 13, 100, "Text",
      CellTags->"IGP_6"],
    Cell[7825, 213, 116, 3, 34, "Text",
      CellTags->"IGP_6"],
    Cell[7944, 218, 470, 14, 40, "Input",
      CellTags->"IGP_6"],
    Cell[8417, 234, 78, 1, 34, "Text",
      CellTags->"IGP_6"],
    Cell[8498, 237, 774, 21, 66, "Input",
      CellTags->"IGP_6"],
    Cell[9275, 260, 198, 4, 56, "Text",
      CellTags->"IGP_6"],
    Cell[9476, 266, 681, 20, 40, "Input",
      CellTags->"IGP_6"],
    Cell[10160, 288, 418, 7, 78, "Text",
      CellTags->"IGP_6"],
    Cell[10581, 297, 1768, 39, 248, "Input",
      CellTags->"IGP_6"],
    Cell[12352, 338, 353, 9, 56, "Text",
      CellTags->"IGP_6"],
    Cell[12730, 351, 505, 13, 92, "Input",
      CellTags->"IGP_6"],
    Cell[13238, 366, 345, 8, 32, "Print",
      CellTags->"IGP_6"],
    Cell[13586, 376, 1019, 23, 57, "Print",
      CellTags->"IGP_6"],
    Cell[14620, 402, 744, 12, 144, "Text",
      CellTags->"IGP_6"],
    Cell[15367, 416, 323, 7, 56, "Text",
      CellTags->"IGP_6"],
    Cell[15693, 425, 6019, 159, 690, "Text",
      CellTags->"IGP_6"],
    Cell[23975, 646, 133, 3, 40, "Input",
      CellTags->"IGP_6"],
    Cell[49357, 1286, 133, 3, 40, "Input",
      CellTags->"IGP_6"],
    Cell[72865, 1854, 133, 3, 40, "Input",
      CellTags->"IGP_6"]},
  "IGP_7"->{
    Cell[21998, 597, 54, 1, 82, "Section",
      CellTags->"IGP_7"]},
  "IGP_7a"->{
    Cell[22077, 602, 59, 1, 50, "Subsubsection",
      CellTags->"IGP_7a"],
    Cell[22139, 605, 1833, 39, 240, "Text",
      CellTags->"IGP_7a"],
    Cell[24111, 651, 7271, 152, 430, "Input",
      CellTags->"IGP_7a"],
    Cell[31385, 805, 317, 8, 40, "Input",
      CellTags->"IGP_7a"],
    Cell[31705, 815, 256, 9, 34, "Text",
      CellTags->"IGP_7a"],
    Cell[31964, 826, 405, 11, 56, "Text",
      CellTags->"IGP_7a"],
    Cell[32372, 839, 407, 7, 78, "Text",
      CellTags->"IGP_7a"],
    Cell[32782, 848, 875, 25, 110, "Text",
      CellTags->"IGP_7a"]},
  "IGP_8"->{
    Cell[33694, 878, 59, 1, 50, "Subsubsection",
      CellTags->"IGP_8"],
    Cell[33756, 881, 1027, 23, 144, "Text",
      CellTags->"IGP_8"],
    Cell[34786, 906, 250, 7, 40, "Input",
      CellTags->"IGP_8"],
    Cell[35039, 915, 338, 7, 66, "Input",
      CellTags->"IGP_8"],
    Cell[35380, 924, 364, 11, 56, "Text",
      CellTags->"IGP_8"],
    Cell[35747, 937, 232, 5, 56, "Text",
      CellTags->"IGP_8"],
    Cell[35982, 944, 490, 15, 56, "Text",
      CellTags->"IGP_8"],
    Cell[36475, 961, 209, 5, 56, "Text",
      CellTags->"IGP_8"]},
  "IGP_9"->{
    Cell[36721, 971, 72, 1, 50, "Subsubsection",
      CellTags->"IGP_9"],
    Cell[36796, 974, 444, 8, 78, "Text",
      CellTags->"IGP_9"],
    Cell[37243, 984, 276, 8, 56, "Text",
      CellTags->"IGP_9"]},
  "IGP_10"->{
    Cell[37568, 998, 65, 1, 82, "Section",
      CellTags->"IGP_10"],
    Cell[37636, 1001, 1951, 32, 298, "Text",
      CellTags->"IGP_10"],
    Cell[39590, 1035, 759, 23, 110, "Text",
      CellTags->"IGP_10"]},
  "IGP_10X"->{
    Cell[40386, 1063, 55, 1, 82, "Section",
      CellTags->"IGP_10X"],
    Cell[40444, 1066, 881, 17, 302, "Text",
      CellTags->"IGP_10X"],
    Cell[41328, 1085, 639, 16, 40, "Input",
      CellTags->"IGP_10X"]},
  "IGP_11"->{
    Cell[42004, 1106, 60, 1, 82, "Section",
      CellTags->"IGP_11"],
    Cell[42067, 1109, 838, 17, 144, "Text",
      CellTags->"IGP_11"],
    Cell[42908, 1128, 1296, 24, 256, "Text",
      CellTags->"IGP_11"]},
  "IGP_12"->{
    Cell[44241, 1157, 56, 1, 82, "Section",
      CellTags->"IGP_12"],
    Cell[44300, 1160, 165, 4, 34, "Text",
      CellTags->"IGP_12"],
    Cell[44468, 1166, 3198, 79, 342, "Text",
      CellTags->"IGP_12"]},
  "IGP_13"->{
    Cell[48284, 1262, 61, 1, 82, "Section",
      CellTags->"IGP_13"]},
  "IGP_14"->{
    Cell[48370, 1267, 87, 1, 50, "Subsubsection",
      CellTags->"IGP_14"],
    Cell[48460, 1270, 894, 14, 182, "Text",
      CellTags->"IGP_14"],
    Cell[49493, 1291, 20983, 493, 1418, "Input",
      CellTags->"IGP_14"],
    Cell[70501, 1788, 594, 14, 92, "Input",
      CellTags->"IGP_14"],
    Cell[71098, 1804, 381, 9, 32, "Print",
      CellTags->"IGP_14"],
    Cell[71482, 1815, 382, 9, 32, "Print",
      CellTags->"IGP_14"],
    Cell[71867, 1826, 385, 9, 32, "Print",
      CellTags->"IGP_14"]},
  "IGP_15"->{
    Cell[72301, 1841, 75, 1, 50, "Subsubsection",
      CellTags->"IGP_15"],
    Cell[72379, 1844, 483, 8, 100, "Text",
      CellTags->"IGP_15"],
    Cell[73001, 1859, 1764, 43, 222, "Input",
      CellTags->"IGP_15"],
    Cell[74768, 1904, 444, 10, 66, "Input",
      CellTags->"IGP_15"],
    Cell[75215, 1916, 567, 12, 100, "Text",
      CellTags->"IGP_15"]},
  "IGP_16"->{
    Cell[75831, 1934, 60, 1, 82, "Section",
      CellTags->"IGP_16"],
    Cell[75894, 1937, 183, 4, 56, "Text",
      CellTags->"IGP_16"],
    Cell[76080, 1943, 200, 5, 64, "Text",
      CellTags->"IGP_16"],
    Cell[76283, 1950, 104, 3, 34, "Text",
      CellTags->"IGP_16"],
    Cell[76390, 1955, 348, 6, 78, "Text",
      CellTags->"IGP_16"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"IGP_1", 77874, 1995},
  {"IGP_3", 78093, 2002},
  {"IGP_4", 78175, 2005},
  {"IGP_5", 78331, 2010},
  {"IGP_6", 78551, 2017},
  {"IGP_7", 80065, 2062},
  {"IGP_7a", 80149, 2065},
  {"IGP_8", 80727, 2082},
  {"IGP_9", 81292, 2099},
  {"IGP_10", 81516, 2106},
  {"IGP_10X", 81745, 2113},
  {"IGP_11", 81976, 2120},
  {"IGP_12", 82205, 2127},
  {"IGP_13", 82432, 2134},
  {"IGP_14", 82518, 2137},
  {"IGP_15", 83038, 2152},
  {"IGP_16", 83415, 2163}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 103, 1, 130, "Chapter"],

Cell[CellGroupData[{
Cell[1904, 58, 88, 2, 82, "Section",
  CellTags->"IGP_1"],

Cell[CellGroupData[{
Cell[2017, 64, 70, 1, 58, "Subsection",
  CellTags->"IGP_1"],
Cell[2090, 67, 3207, 62, 454, "Text",
  CellTags->"IGP_1"],
Cell[5300, 131, 85, 2, 34, "Text"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[5434, 139, 68, 1, 82, "Section",
  CellTags->"IGP_3"],

Cell[CellGroupData[{
Cell[5527, 144, 68, 1, 50, "Subsubsection",
  CellTags->"IGP_4"],
Cell[5598, 147, 985, 21, 152, "Text",
  CellTags->"IGP_4"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6620, 173, 65, 1, 50, "Subsubsection",
  CellTags->"IGP_5"],
Cell[6688, 176, 89, 1, 34, "Text",
  CellTags->"IGP_5"],
Cell[6780, 179, 120, 3, 40, "Input",
  CellTags->"IGP_5"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6937, 187, 56, 1, 50, "Subsubsection",
  CellTags->"IGP_6"],
Cell[6996, 190, 72, 1, 34, "Text",
  CellTags->"IGP_6"],
Cell[7071, 193, 133, 3, 40, "Input",
  CellTags->"IGP_6"],
Cell[7207, 198, 615, 13, 100, "Text",
  CellTags->"IGP_6"],
Cell[7825, 213, 116, 3, 34, "Text",
  CellTags->"IGP_6"],
Cell[7944, 218, 470, 14, 40, "Input",
  CellTags->"IGP_6"],
Cell[8417, 234, 78, 1, 34, "Text",
  CellTags->"IGP_6"],
Cell[8498, 237, 774, 21, 66, "Input",
  CellTags->"IGP_6"],
Cell[9275, 260, 198, 4, 56, "Text",
  CellTags->"IGP_6"],
Cell[9476, 266, 681, 20, 40, "Input",
  CellTags->"IGP_6"],
Cell[10160, 288, 418, 7, 78, "Text",
  CellTags->"IGP_6"],
Cell[10581, 297, 1768, 39, 248, "Input",
  CellTags->"IGP_6"],
Cell[12352, 338, 353, 9, 56, "Text",
  CellTags->"IGP_6"],

Cell[CellGroupData[{
Cell[12730, 351, 505, 13, 92, "Input",
  CellTags->"IGP_6"],
Cell[13238, 366, 345, 8, 32, "Print",
  CellTags->"IGP_6"],
Cell[13586, 376, 1019, 23, 57, "Print",
  CellTags->"IGP_6"]
}, Open  ]],
Cell[14620, 402, 744, 12, 144, "Text",
  CellTags->"IGP_6"],
Cell[15367, 416, 323, 7, 56, "Text",
  CellTags->"IGP_6"],
Cell[15693, 425, 6019, 159, 690, "Text",
  CellTags->"IGP_6"],
Cell[21715, 586, 234, 5, 56, "Text"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[21998, 597, 54, 1, 82, "Section",
  CellTags->"IGP_7"],

Cell[CellGroupData[{
Cell[22077, 602, 59, 1, 50, "Subsubsection",
  CellTags->"IGP_7a"],
Cell[22139, 605, 1833, 39, 240, "Text",
  CellTags->"IGP_7a"],
Cell[23975, 646, 133, 3, 40, "Input",
  CellTags->"IGP_6"],
Cell[24111, 651, 7271, 152, 430, "Input",
  CellTags->"IGP_7a"],
Cell[31385, 805, 317, 8, 40, "Input",
  CellTags->"IGP_7a"],
Cell[31705, 815, 256, 9, 34, "Text",
  CellTags->"IGP_7a"],
Cell[31964, 826, 405, 11, 56, "Text",
  CellTags->"IGP_7a"],
Cell[32372, 839, 407, 7, 78, "Text",
  CellTags->"IGP_7a"],
Cell[32782, 848, 875, 25, 110, "Text",
  CellTags->"IGP_7a"]
}, Open  ]],

Cell[CellGroupData[{
Cell[33694, 878, 59, 1, 50, "Subsubsection",
  CellTags->"IGP_8"],
Cell[33756, 881, 1027, 23, 144, "Text",
  CellTags->"IGP_8"],
Cell[34786, 906, 250, 7, 40, "Input",
  CellTags->"IGP_8"],
Cell[35039, 915, 338, 7, 66, "Input",
  CellTags->"IGP_8"],
Cell[35380, 924, 364, 11, 56, "Text",
  CellTags->"IGP_8"],
Cell[35747, 937, 232, 5, 56, "Text",
  CellTags->"IGP_8"],
Cell[35982, 944, 490, 15, 56, "Text",
  CellTags->"IGP_8"],
Cell[36475, 961, 209, 5, 56, "Text",
  CellTags->"IGP_8"]
}, Open  ]],

Cell[CellGroupData[{
Cell[36721, 971, 72, 1, 50, "Subsubsection",
  CellTags->"IGP_9"],
Cell[36796, 974, 444, 8, 78, "Text",
  CellTags->"IGP_9"],
Cell[37243, 984, 276, 8, 56, "Text",
  CellTags->"IGP_9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[37568, 998, 65, 1, 82, "Section",
  CellTags->"IGP_10"],
Cell[37636, 1001, 1951, 32, 298, "Text",
  CellTags->"IGP_10"],
Cell[39590, 1035, 759, 23, 110, "Text",
  CellTags->"IGP_10"]
}, Open  ]],

Cell[CellGroupData[{
Cell[40386, 1063, 55, 1, 82, "Section",
  CellTags->"IGP_10X"],
Cell[40444, 1066, 881, 17, 302, "Text",
  CellTags->"IGP_10X"],
Cell[41328, 1085, 639, 16, 40, "Input",
  CellTags->"IGP_10X"]
}, Open  ]],

Cell[CellGroupData[{
Cell[42004, 1106, 60, 1, 82, "Section",
  CellTags->"IGP_11"],
Cell[42067, 1109, 838, 17, 144, "Text",
  CellTags->"IGP_11"],
Cell[42908, 1128, 1296, 24, 256, "Text",
  CellTags->"IGP_11"]
}, Open  ]],

Cell[CellGroupData[{
Cell[44241, 1157, 56, 1, 82, "Section",
  CellTags->"IGP_12"],
Cell[44300, 1160, 165, 4, 34, "Text",
  CellTags->"IGP_12"],
Cell[44468, 1166, 3198, 79, 342, "Text",
  CellTags->"IGP_12"],
Cell[47669, 1247, 578, 10, 100, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[48284, 1262, 61, 1, 82, "Section",
  CellTags->"IGP_13"],

Cell[CellGroupData[{
Cell[48370, 1267, 87, 1, 50, "Subsubsection",
  CellTags->"IGP_14"],
Cell[48460, 1270, 894, 14, 182, "Text",
  CellTags->"IGP_14"],
Cell[49357, 1286, 133, 3, 40, "Input",
  CellTags->"IGP_6"],
Cell[49493, 1291, 20983, 493, 1418, "Input",
  CellTags->"IGP_14"],

Cell[CellGroupData[{
Cell[70501, 1788, 594, 14, 92, "Input",
  CellTags->"IGP_14"],
Cell[71098, 1804, 381, 9, 32, "Print",
  CellTags->"IGP_14"],
Cell[71482, 1815, 382, 9, 32, "Print",
  CellTags->"IGP_14"],
Cell[71867, 1826, 385, 9, 32, "Print",
  CellTags->"IGP_14"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[72301, 1841, 75, 1, 50, "Subsubsection",
  CellTags->"IGP_15"],
Cell[72379, 1844, 483, 8, 100, "Text",
  CellTags->"IGP_15"],
Cell[72865, 1854, 133, 3, 40, "Input",
  CellTags->"IGP_6"],
Cell[73001, 1859, 1764, 43, 222, "Input",
  CellTags->"IGP_15"],
Cell[74768, 1904, 444, 10, 66, "Input",
  CellTags->"IGP_15"],
Cell[75215, 1916, 567, 12, 100, "Text",
  CellTags->"IGP_15"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[75831, 1934, 60, 1, 82, "Section",
  CellTags->"IGP_16"],
Cell[75894, 1937, 183, 4, 56, "Text",
  CellTags->"IGP_16"],
Cell[76080, 1943, 200, 5, 64, "Text",
  CellTags->"IGP_16"],
Cell[76283, 1950, 104, 3, 34, "Text",
  CellTags->"IGP_16"],
Cell[76390, 1955, 348, 6, 78, "Text",
  CellTags->"IGP_16"]
}, Open  ]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

